@@ -1,4 +1,4 @@
-/*	$OpenBSD: bgpctl.c,v 1.76 2005/03/14 12:36:27 henning Exp $ */
+/*	$OpenBSD: bgpctl.c,v 1.100 2006/01/24 15:28:03 henning Exp $ */
 
 /*
  * Copyright (c) 2003 Henning Brauer <henning@openbsd.org>
@@ -29,6 +29,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <util.h>
 
 #include "bgpd.h"
 #include "session.h"
@@ -46,6 +47,7 @@ int		 main(int, char *[]);
 char		*fmt_peer(const struct peer_config *, int);
 void		 show_summary_head(void);
 int		 show_summary_msg(struct imsg *, int);
+int		 show_summary_terse_msg(struct imsg *, int);
 int		 show_neighbor_msg(struct imsg *, enum neighbor_views);
 void		 print_neighbor_capa_mp_safi(u_int8_t);
 void		 print_neighbor_msgstats(struct peer *);
@@ -55,6 +57,7 @@ static char	*fmt_timeframe(time_t t);
 static char	*fmt_timeframe_core(time_t t);
 void		 show_fib_head(void);
 void		 show_network_head(void);
+void		 show_fib_flags(u_int16_t);
 int		 show_fib_msg(struct imsg *);
 void		 show_nexthop_head(void);
 int		 show_nexthop_msg(struct imsg *);
@@ -68,6 +71,14 @@ void		 show_rib_summary_head(void);
 void		 print_prefix(struct bgpd_addr *, u_int8_t, u_int8_t);
 const char *	 print_origin(u_int8_t, int);
 int		 show_rib_summary_msg(struct imsg *);
+char		*fmt_mem(int64_t);
+int		 show_rib_memory_msg(struct imsg *);
+void		 send_filterset(struct imsgbuf *, struct filter_set_head *);
+static const char	*get_errstr(u_int8_t, u_int8_t);
+int		 show_result(struct imsg *);
+void		 log_warnx(const char *, ...);
+void		 log_warn(const char *, ...);
+void		 fatal(const char *);
 
 struct imsgbuf	*ibuf;
 
@@ -76,7 +87,8 @@ usage(void)
 {
 	extern char	*__progname;
 
-	fprintf(stderr, "usage: %s [-n] <command> [arg [...]]\n", __progname);
+	fprintf(stderr, "usage: %s [-s socket] [-n] "
+	    "<command> [arg [...]]\n", __progname);
 	exit(1);
 }
 
@@ -89,13 +101,18 @@ main(int argc, char *argv[])
 	struct network_config	 net;
 	struct parse_result	*res;
 	struct ctl_neighbor	 neighbor;
+	char			*sockname;
 
-	while ((ch = getopt(argc, argv, "n")) != -1) {
+	sockname = SOCKET_NAME;
+	while ((ch = getopt(argc, argv, "ns:")) != -1) {
 		switch (ch) {
 		case 'n':
 			if (++nodescr > 1)
 				usage();
 			break;
+		case 's':
+			sockname = optarg;
+			break;
 		default:
 			usage();
 			/* NOTREACHED */
@@ -115,12 +132,14 @@ main(int argc, char *argv[])
 
 	bzero(&sun, sizeof(sun));
 	sun.sun_family = AF_UNIX;
-	strlcpy(sun.sun_path, SOCKET_NAME, sizeof(sun.sun_path));
+	if (strlcpy(sun.sun_path, sockname, sizeof(sun.sun_path)) >=
+	    sizeof(sun.sun_path))
+		errx(1, "socket name too long");
 	if (connect(fd, (struct sockaddr *)&sun, sizeof(sun)) == -1)
-		err(1, "connect: %s", SOCKET_NAME);
+		err(1, "connect: %s", sockname);
 
 	if ((ibuf = malloc(sizeof(struct imsgbuf))) == NULL)
-		fatal(NULL);
+		err(1, NULL);
 	imsg_init(ibuf, fd);
 	done = 0;
 
@@ -133,11 +152,23 @@ main(int argc, char *argv[])
 		imsg_compose(ibuf, IMSG_CTL_SHOW_NEIGHBOR, 0, 0, -1, NULL, 0);
 		show_summary_head();
 		break;
+	case SHOW_SUMMARY_TERSE:
+		imsg_compose(ibuf, IMSG_CTL_SHOW_TERSE, 0, 0, -1, NULL, 0);
+		break;
 	case SHOW_FIB:
-		if (!res->addr.af)
-			imsg_compose(ibuf, IMSG_CTL_KROUTE, 0, 0, -1,
-			    &res->flags, sizeof(res->flags));
-		else
+		if (!res->addr.af) {
+			struct buf	*msg;
+
+			if ((msg = imsg_create(ibuf, IMSG_CTL_KROUTE, 0, 0,
+			    sizeof(res->flags) + sizeof(res->af))) == NULL)
+				errx(1, "imsg_create failure");
+			if (imsg_add(msg, &res->flags, sizeof(res->flags)) ==
+			    -1 ||
+			    imsg_add(msg, &res->af, sizeof(res->af)) == -1)
+				errx(1, "imsg_add failure");
+			if (imsg_close(ibuf, msg) < 0)
+				errx(1, "imsg_close error");
+		} else
 			imsg_compose(ibuf, IMSG_CTL_KROUTE_ADDR, 0, 0, -1,
 			    &res->addr, sizeof(res->addr));
 		show_fib_head();
@@ -174,9 +205,12 @@ main(int argc, char *argv[])
 			    &msg, sizeof(msg));
 		} else
 			imsg_compose(ibuf, IMSG_CTL_SHOW_RIB, 0, 0, -1,
-			    NULL, 0);
+			    &res->af, sizeof(res->af));
 		show_rib_summary_head();
 		break;
+	case SHOW_RIB_MEM:
+		imsg_compose(ibuf, IMSG_CTL_SHOW_RIB_MEM, 0, 0, -1, NULL, 0);
+		break;
 	case RELOAD:
 		imsg_compose(ibuf, IMSG_CTL_RELOAD, 0, 0, -1, NULL, 0);
 		printf("reload request sent.\n");
@@ -201,20 +235,14 @@ main(int argc, char *argv[])
 	case NEIGHBOR_UP:
 		imsg_compose(ibuf, IMSG_CTL_NEIGHBOR_UP, 0, 0, -1,
 		    &neighbor, sizeof(neighbor));
-		printf("request sent.\n");
-		done = 1;
 		break;
 	case NEIGHBOR_DOWN:
 		imsg_compose(ibuf, IMSG_CTL_NEIGHBOR_DOWN, 0, 0, -1,
 		    &neighbor, sizeof(neighbor));
-		printf("request sent.\n");
-		done = 1;
 		break;
 	case NEIGHBOR_CLEAR:
 		imsg_compose(ibuf, IMSG_CTL_NEIGHBOR_CLEAR, 0, 0, -1,
 		    &neighbor, sizeof(neighbor));
-		printf("request sent.\n");
-		done = 1;
 		break;
 	case NETWORK_ADD:
 	case NETWORK_REMOVE:
@@ -225,6 +253,7 @@ main(int argc, char *argv[])
 		if (res->action == NETWORK_ADD) {
 			imsg_compose(ibuf, IMSG_NETWORK_ADD, 0, 0, -1,
 			    &net, sizeof(net));
+			send_filterset(ibuf, &res->set);
 			imsg_compose(ibuf, IMSG_NETWORK_DONE, 0, 0, -1,
 			    NULL, 0);
 		} else
@@ -239,7 +268,8 @@ main(int argc, char *argv[])
 		done = 1;
 		break;
 	case NETWORK_SHOW:
-		imsg_compose(ibuf, IMSG_CTL_SHOW_NETWORK, 0, 0, -1, NULL, 0);
+		imsg_compose(ibuf, IMSG_CTL_SHOW_NETWORK, 0, 0, -1,
+		    &res->af, sizeof(res->af));
 		show_network_head();
 		break;
 	}
@@ -259,11 +289,21 @@ main(int argc, char *argv[])
 				errx(1, "imsg_get error");
 			if (n == 0)
 				break;
+
+			if (imsg.hdr.type == IMSG_CTL_RESULT) {
+				done = show_result(&imsg);
+				imsg_free(&imsg);
+				continue;
+			}
+
 			switch (res->action) {
 			case SHOW:
 			case SHOW_SUMMARY:
 				done = show_summary_msg(&imsg, nodescr);
 				break;
+			case SHOW_SUMMARY_TERSE:
+				done = show_summary_terse_msg(&imsg, nodescr);
+				break;
 			case SHOW_FIB:
 				done = show_fib_msg(&imsg);
 				break;
@@ -282,18 +322,21 @@ main(int argc, char *argv[])
 			case SHOW_RIB:
 				done = show_rib_summary_msg(&imsg);
 				break;
+			case SHOW_RIB_MEM:
+				done = show_rib_memory_msg(&imsg);
+				break;
 			case NETWORK_SHOW:
 				done = show_fib_msg(&imsg);
 				break;
+			case NEIGHBOR:
+			case NEIGHBOR_UP:
+			case NEIGHBOR_DOWN:
+			case NEIGHBOR_CLEAR:
 			case NONE:
 			case RELOAD:
 			case FIB:
 			case FIB_COUPLE:
 			case FIB_DECOUPLE:
-			case NEIGHBOR:
-			case NEIGHBOR_UP:
-			case NEIGHBOR_DOWN:
-			case NEIGHBOR_CLEAR:
 			case NETWORK_ADD:
 			case NETWORK_REMOVE:
 			case NETWORK_FLUSH:
@@ -316,7 +359,7 @@ fmt_peer(const struct peer_config *peer, int nodescr)
 
 	if (peer->descr[0] && !nodescr) {
 		if ((p = strdup(peer->descr)) == NULL)
-			fatal(NULL);
+			err(1, NULL);
 		return (p);
 	}
 
@@ -324,10 +367,10 @@ fmt_peer(const struct peer_config *peer, int nodescr)
 	if ((peer->remote_addr.af == AF_INET && peer->remote_masklen != 32) ||
 	    (peer->remote_addr.af == AF_INET6 && peer->remote_masklen != 128)) {
 		if (asprintf(&p, "%s/%u", ip, peer->remote_masklen) == -1)
-			fatal(NULL);
+			err(1, NULL);
 	} else {
 		if ((p = strdup(ip)) == NULL)
-			fatal(NULL);
+			err(1, NULL);
 	}
 
 	return (p);
@@ -380,6 +423,29 @@ show_summary_msg(struct imsg *imsg, int nodescr)
 	return (0);
 }
 
+int
+show_summary_terse_msg(struct imsg *imsg, int nodescr)
+{
+	struct peer		*p;
+	char			*s;
+
+	switch (imsg->hdr.type) {
+	case IMSG_CTL_SHOW_NEIGHBOR:
+		p = imsg->data;
+		s = fmt_peer(&p->conf, nodescr);
+		printf("%s %u %s\n", s, p->conf.remote_as,
+		    statenames[p->state]);
+		free(s);
+		break;
+	case IMSG_CTL_END:
+		return (1);
+	default:
+		break;
+	}
+
+	return (0);
+}
+
 int
 show_neighbor_msg(struct imsg *imsg, enum neighbor_views nv)
 {
@@ -399,7 +465,8 @@ show_neighbor_msg(struct imsg *imsg, enum neighbor_views nv)
 			    p->conf.remote_masklen) == -1)
 				err(1, NULL);
 		} else
-			s = strdup(log_addr(&p->conf.remote_addr));
+			if ((s = strdup(log_addr(&p->conf.remote_addr))) == NULL)
+				err(1, "strdup");
 
 		ina.s_addr = p->remote_bgpid;
 		printf("BGP neighbor is %s, ", s);
@@ -450,24 +517,35 @@ show_neighbor_msg(struct imsg *imsg, enum neighbor_views nv)
 			break;
 		}
 		printf("\n");
-		if (getnameinfo((struct sockaddr *)&p->sa_local,
-		    (socklen_t)p->sa_local.ss_len,
-		    buf, sizeof(buf), pbuf, sizeof(pbuf),
-		    NI_NUMERICHOST | NI_NUMERICSERV)) {
-			strlcpy(buf, "(unknown)", sizeof(buf));
-			strlcpy(pbuf, "", sizeof(pbuf));
-		}
-		printf("  Local host:  %20s, Local port:  %5s\n", buf, pbuf);
+		if (p->state == STATE_IDLE) {
+			static const char	*errstr;
 
-		if (getnameinfo((struct sockaddr *)&p->sa_remote,
-		    (socklen_t)p->sa_remote.ss_len,
-		    buf, sizeof(buf), pbuf, sizeof(pbuf),
-		    NI_NUMERICHOST | NI_NUMERICSERV)) {
-			strlcpy(buf, "(unknown)", sizeof(buf));
-			strlcpy(pbuf, "", sizeof(pbuf));
+			errstr = get_errstr(p->stats.last_sent_errcode,
+			    p->stats.last_sent_suberr);
+			if (errstr)
+				printf("  Last error: %s\n\n", errstr);
+		} else {
+			if (getnameinfo((struct sockaddr *)&p->sa_local,
+			    (socklen_t)p->sa_local.ss_len,
+			    buf, sizeof(buf), pbuf, sizeof(pbuf),
+			    NI_NUMERICHOST | NI_NUMERICSERV)) {
+				strlcpy(buf, "(unknown)", sizeof(buf));
+				strlcpy(pbuf, "", sizeof(pbuf));
+			}
+			printf("  Local host:  %20s, Local port:  %5s\n", buf,
+			    pbuf);
+
+			if (getnameinfo((struct sockaddr *)&p->sa_remote,
+			    (socklen_t)p->sa_remote.ss_len,
+			    buf, sizeof(buf), pbuf, sizeof(pbuf),
+			    NI_NUMERICHOST | NI_NUMERICSERV)) {
+				strlcpy(buf, "(unknown)", sizeof(buf));
+				strlcpy(pbuf, "", sizeof(pbuf));
+			}
+			printf("  Remote host: %20s, Remote port: %5s\n", buf,
+			    pbuf);
+			printf("\n");
 		}
-		printf("  Remote host: %20s, Remote port: %5s\n", buf, pbuf);
-		printf("\n");
 		break;
 	case IMSG_CTL_END:
 		return (1);
@@ -600,7 +678,8 @@ void
 show_fib_head(void)
 {
 	printf("flags: * = valid, B = BGP, C = Connected, S = Static\n");
-	printf("       N = BGP Nexthop reachable via this route\n\n");
+	printf("       N = BGP Nexthop reachable via this route\n");
+	printf("       r = reject route, b = blackhole route\n\n");
 	printf("flags destination          gateway\n");
 }
 
@@ -611,10 +690,45 @@ show_network_head(void)
 	printf("flags destination\n");
 }
 
+void
+show_fib_flags(u_int16_t flags)
+{
+	if (flags & F_DOWN)
+		printf(" ");
+	else
+		printf("*");
+
+	if (flags & F_BGPD_INSERTED)
+		printf("B");
+	else if (flags & F_CONNECTED)
+		printf("C");
+	else if (flags & F_STATIC)
+		printf("S");
+	else
+		printf(" ");
+
+	if (flags & F_NEXTHOP)
+		printf("N");
+	else
+		printf(" ");
+
+	if (flags & F_REJECT && flags & F_BLACKHOLE)
+		printf("f");
+	else if (flags & F_REJECT)
+		printf("r");
+	else if (flags & F_BLACKHOLE)
+		printf("b");
+	else
+		printf(" ");
+
+	printf("  ");
+}
+
 int
 show_fib_msg(struct imsg *imsg)
 {
 	struct kroute		*k;
+	struct kroute6		*k6;
 	char			*p;
 
 	switch (imsg->hdr.type) {
@@ -624,26 +738,8 @@ show_fib_msg(struct imsg *imsg)
 			errx(1, "wrong imsg len");
 		k = imsg->data;
 
-		if (k->flags & F_DOWN)
-			printf(" ");
-		else
-			printf("*");
+		show_fib_flags(k->flags);
 
-		if (k->flags & F_BGPD_INSERTED)
-			printf("B");
-		else if (k->flags & F_CONNECTED)
-			printf("C");
-		else if (k->flags & F_KERNEL)
-			printf("S");
-		else
-			printf(" ");
-
-		if (k->flags & F_NEXTHOP)
-			printf("N");
-		else
-			printf(" ");
-
-		printf("   ");
 		if (asprintf(&p, "%s/%u", inet_ntoa(k->prefix), k->prefixlen) ==
 		    -1)
 			err(1, NULL);
@@ -656,6 +752,27 @@ show_fib_msg(struct imsg *imsg)
 			printf("link#%u", k->ifindex);
 		printf("\n");
 
+		break;
+	case IMSG_CTL_KROUTE6:
+	case IMSG_CTL_SHOW_NETWORK6:
+		if (imsg->hdr.len < IMSG_HEADER_SIZE + sizeof(struct kroute6))
+			errx(1, "wrong imsg len");
+		k6 = imsg->data;
+
+		show_fib_flags(k6->flags);
+
+		if (asprintf(&p, "%s/%u", log_in6addr(&k6->prefix),
+		    k6->prefixlen) == -1)
+			err(1, NULL);
+		printf("%-20s ", p);
+		free(p);
+
+		if (!IN6_IS_ADDR_UNSPECIFIED(&k6->nexthop))
+			printf("%s", log_in6addr(&k6->nexthop));
+		else if (k6->flags & F_CONNECTED)
+			printf("link#%u", k6->ifindex);
+		printf("\n");
+
 		break;
 	case IMSG_CTL_END:
 		return (1);
@@ -682,7 +799,7 @@ show_nexthop_msg(struct imsg *imsg)
 	switch (imsg->hdr.type) {
 	case IMSG_CTL_SHOW_NEXTHOP:
 		p = imsg->data;
-		printf("%-20s %-10s", inet_ntoa(p->addr.v4),
+		printf("%-20s %-10s", log_addr(&p->addr),
 		    p->valid ? "valid" : "invalid");
 		if (p->kif.ifname[0]) {
 			printf("%-8s", p->kif.ifname);
@@ -833,7 +950,7 @@ show_rib_summary_head(void)
 	printf(
 	    "flags: * = Valid, > = Selected, I = via IBGP, A = Announced\n");
 	printf("origin: i = IGP, e = EGP, ? = Incomplete\n\n");
-	printf("%-4s %-20s%-15s  %5s %5s %s\n", "flags", "destination",
+	printf("%-5s %-20s%-15s  %5s %5s %s\n", "flags", "destination",
 	    "gateway", "lpref", "med", "aspath origin");
 }
 
@@ -854,9 +971,9 @@ print_prefix(struct bgpd_addr *prefix, u_int8_t prefixlen, u_int8_t flags)
 		*p++ = '>';
 	*p = '\0';
 
-	if (asprintf(&p, "%s/%u", inet_ntoa(prefix->v4), prefixlen) == -1)
+	if (asprintf(&p, "%s/%u", log_addr(prefix), prefixlen) == -1)
 		err(1, NULL);
-	printf("%-4s %-20s", flagstr, p);
+	printf("%-5s %-20s", flagstr, p);
 	free(p);
 }
 
@@ -875,84 +992,191 @@ print_origin(u_int8_t origin, int sum)
 	}
 }
 
-char			*aspath = NULL;
-struct ctl_show_rib	*rib = NULL;
-
 int
 show_rib_summary_msg(struct imsg *imsg)
 {
-	struct ctl_show_rib_prefix	*p;
-	u_char				*asdata;
+	struct ctl_show_rib	 rib;
+	char			*aspath;
+	u_char			*asdata;
 
 	switch (imsg->hdr.type) {
 	case IMSG_CTL_SHOW_RIB:
-		if (rib != NULL) {
-			free(rib);
-			rib = NULL;
-		}
-		if (aspath != NULL) {
-			free(aspath);
-			aspath = NULL;
-		}
+		memcpy(&rib, imsg->data, sizeof(rib));
 
-		if ((rib = malloc(imsg->hdr.len - IMSG_HEADER_SIZE)) == NULL)
-			err(1, NULL);
-		memcpy(rib, imsg->data, imsg->hdr.len - IMSG_HEADER_SIZE);
+		print_prefix(&rib.prefix, rib.prefixlen, rib.flags);
+		printf("%-15s ", log_addr(&rib.nexthop));
 
-		print_prefix(&rib->prefix, rib->prefixlen, rib->flags);
-		printf("%-15s ", inet_ntoa(rib->nexthop.v4));
-
-		printf(" %5u %5u ", rib->local_pref, rib->med);
+		printf(" %5u %5u ", rib.local_pref, rib.med);
 
 		asdata = imsg->data;
 		asdata += sizeof(struct ctl_show_rib);
-		if (aspath_asprint(&aspath, asdata, rib->aspath_len) == -1)
+		if (aspath_asprint(&aspath, asdata, rib.aspath_len) == -1)
 			err(1, NULL);
 		if (strlen(aspath) > 0)
 			printf("%s ", aspath);
+		free(aspath);
 
-		printf("%s\n", print_origin(rib->origin, 1));
-		break;
-	case IMSG_CTL_SHOW_RIB_PREFIX:
-		p = imsg->data;
-		if (rib == NULL)
-			/* unexpected packet */
-			return (0);
-
-		print_prefix(&p->prefix, p->prefixlen, p->flags);
-		printf("%-15s ", inet_ntoa(rib->nexthop.v4));
-
-		printf(" %5u %5u ", rib->local_pref, rib->med);
-
-		if (strlen(aspath) > 0)
-			printf("%s ", aspath);
-
-		printf("%s\n", print_origin(rib->origin, 1));
+		printf("%s\n", print_origin(rib.origin, 1));
 		break;
 	case IMSG_CTL_END:
-		if (rib != NULL) {
-			free(rib);
-			rib = NULL;
-		}
-		if (aspath != NULL) {
-			free(aspath);
-			aspath = NULL;
-		}
-
 		return (1);
 	default:
-		if (rib != NULL) {
-			free(rib);
-			rib = NULL;
-		}
-		if (aspath != NULL) {
-			free(aspath);
-			aspath = NULL;
-		}
-
 		break;
 	}
 
 	return (0);
 }
 
+char *
+fmt_mem(int64_t num)
+{
+	static char	buf[16];
+
+	if (fmt_scaled(num, buf) == -1)
+		snprintf(buf, sizeof(buf), "%lldB", num);
+
+	return (buf);
+}
+
+int
+show_rib_memory_msg(struct imsg *imsg)
+{
+	struct rde_memstats	stats;
+
+	switch (imsg->hdr.type) {
+	case IMSG_CTL_SHOW_RIB_MEM:
+		memcpy(&stats, imsg->data, sizeof(stats));
+		printf("RDE memory statistics\n");
+		printf("%10lld IPv4 network entries using %s of memory\n",
+		    stats.pt4_cnt, fmt_mem(stats.pt4_cnt *
+		    sizeof(struct pt_entry4)));
+		if (stats.pt6_cnt != 0)
+			printf("%10lld IPv6 network entries using "
+			    "%s of memory\n", stats.pt6_cnt,
+			    fmt_mem(stats.pt6_cnt * sizeof(struct pt_entry6)));
+		printf("%10lld prefix entries using %s of memory\n",
+		    stats.prefix_cnt, fmt_mem(stats.prefix_cnt *
+		    sizeof(struct prefix)));
+		printf("%10lld BGP path attribute entries using %s of memory\n",
+		    stats.path_cnt, fmt_mem(stats.path_cnt *
+		    sizeof(struct rde_aspath)));
+		printf("%10lld BGP AS-PATH attribute entries using "
+		    "%s of memory,\n\t   and holding %lld references\n",
+		    stats.aspath_cnt, fmt_mem(stats.aspath_size),
+		    stats.aspath_refs);
+		printf("%10lld BGP attributes entries using %s of memory\n",
+		    stats.attr_cnt, fmt_mem(stats.attr_cnt *
+		    sizeof(struct attr)));
+		printf("\t   and holding %lld references\n", stats.attr_refs);
+		printf("%10lld BGP attributes using %s of memory\n",
+		    stats.attr_dcnt, fmt_mem(stats.attr_data));
+		printf("RIB using %s of memory\n", fmt_mem(
+		    stats.pt4_cnt * sizeof(struct pt_entry4) +
+		    stats.pt6_cnt * sizeof(struct pt_entry6) +
+		    stats.prefix_cnt * sizeof(struct prefix) +
+		    stats.path_cnt * sizeof(struct rde_aspath) +
+		    stats.aspath_size + stats.attr_cnt * sizeof(struct attr) +
+		    stats.attr_data));
+		break;
+	default:
+		break;
+	}
+
+	return (1);
+}
+
+void
+send_filterset(struct imsgbuf *i, struct filter_set_head *set)
+{
+	struct filter_set	*s;
+
+	while ((s = TAILQ_FIRST(set)) != NULL) {
+		imsg_compose(i, IMSG_FILTER_SET, 0, 0, -1, s,
+		    sizeof(struct filter_set));
+		TAILQ_REMOVE(set, s, entry);
+		free(s);
+	}
+}
+
+static const char *
+get_errstr(u_int8_t errcode, u_int8_t subcode)
+{
+	static const char	*errstr = NULL;
+
+	if (errcode && errcode < sizeof(errnames)/sizeof(char *))
+		errstr = errnames[errcode];
+
+	switch (errcode) {
+	case ERR_HEADER:
+		if (subcode &&
+		    subcode < sizeof(suberr_header_names)/sizeof(char *))
+			errstr = suberr_header_names[subcode];
+		break;
+	case ERR_OPEN:
+		if (subcode &&
+		    subcode < sizeof(suberr_open_names)/sizeof(char *))
+			errstr = suberr_open_names[subcode];
+		break;
+	case ERR_UPDATE:
+		if (subcode &&
+		    subcode < sizeof(suberr_update_names)/sizeof(char *))
+			errstr = suberr_update_names[subcode];
+		break;
+	case ERR_HOLDTIMEREXPIRED:
+	case ERR_FSM:
+	case ERR_CEASE:
+		break;
+	default:
+		return ("unknown error code");
+	}
+
+	return (errstr);
+}
+
+int
+show_result(struct imsg *imsg)
+{
+	u_int	rescode;
+
+	if (imsg->hdr.len != IMSG_HEADER_SIZE + sizeof(rescode))
+		errx(1, "got IMSG_CTL_RESULT with wrong len");
+	memcpy(&rescode, imsg->data, sizeof(rescode));
+
+	if (rescode == 0)
+		printf("request processed\n");
+	else {
+		if (rescode >
+		    sizeof(ctl_res_strerror)/sizeof(ctl_res_strerror[0]))
+			errx(1, "illegal error code %u", rescode);
+		printf("%s\n", ctl_res_strerror[rescode]);
+	}
+
+	return (1);
+}
+
+/* following functions are necessary for imsg framework */
+void
+log_warnx(const char *emsg, ...)
+{
+	va_list	 ap;
+
+	va_start(ap, emsg);
+	vwarnx(emsg, ap);
+	va_end(ap);
+}
+
+void
+log_warn(const char *emsg, ...)
+{
+	va_list	 ap;
+
+	va_start(ap, emsg);
+	vwarn(emsg, ap);
+	va_end(ap);
+}
+
+void
+fatal(const char *emsg)
+{
+	err(1, emsg);
+}
