@@ -1,4 +1,4 @@
-/*	$OpenBSD: bgpd.c,v 1.113 2005/02/09 10:56:28 henning Exp $ */
+/*	$OpenBSD: bgpd.c,v 1.132 2006/01/24 14:26:52 claudio Exp $ */
 
 /*
  * Copyright (c) 2003, 2004 Henning Brauer <henning@openbsd.org>
@@ -41,12 +41,17 @@ __dead void	usage(void);
 int		main(int, char *[]);
 int		check_child(pid_t, const char *);
 int		send_filterset(struct imsgbuf *, struct filter_set_head *,
-		    int, int);
+		    int);
 int		reconfigure(char *, struct bgpd_config *, struct mrt_head *,
 		    struct peer **, struct filter_head *);
 int		dispatch_imsg(struct imsgbuf *, int);
 
 int			 rfd = -1;
+int			 cflags = 0;
+struct filter_set_head	*connectset;
+struct filter_set_head	*connectset6;
+struct filter_set_head	*staticset;
+struct filter_set_head	*staticset6;
 volatile sig_atomic_t	 mrtdump = 0;
 volatile sig_atomic_t	 quit = 0;
 volatile sig_atomic_t	 reconfig = 0;
@@ -81,7 +86,7 @@ usage(void)
 	extern char *__progname;
 
 	fprintf(stderr, "usage: %s [-dnv] ", __progname);
-	fprintf(stderr, "[-D macro=value] [-f file]\n");
+	fprintf(stderr, "[-D macro=value] [-f file] [-r path]\n");
 	exit(1);
 }
 
@@ -107,7 +112,7 @@ main(int argc, char *argv[])
 	pid_t			 io_pid = 0, rde_pid = 0, pid;
 	char			*conffile;
 	int			 debug = 0;
-	int			 ch, nfds, timeout;
+	int			 ch, timeout, nfds;
 	int			 pipe_m2s[2];
 	int			 pipe_m2r[2];
 	int			 pipe_s2r[2];
@@ -126,7 +131,7 @@ main(int argc, char *argv[])
 	TAILQ_INIT(rules_l);
 	peer_l = NULL;
 
-	while ((ch = getopt(argc, argv, "dD:f:nv")) != -1) {
+	while ((ch = getopt(argc, argv, "dD:f:nr:v")) != -1) {
 		switch (ch) {
 		case 'd':
 			debug = 1;
@@ -147,14 +152,19 @@ main(int argc, char *argv[])
 				conf.opts |= BGPD_OPT_VERBOSE2;
 			conf.opts |= BGPD_OPT_VERBOSE;
 			break;
+		case 'r':
+			conf.rcsock = optarg;
+			break;
 		default:
 			usage();
 			/* NOTREACHED */
 		}
 	}
 
-	if (parse_config(conffile, &conf, &mrt_l, &peer_l, &net_l, rules_l))
+	if (parse_config(conffile, &conf, &mrt_l, &peer_l, &net_l, rules_l)) {
+		free(rules_l);
 		exit(1);
+	}
 
 	if (conf.opts & BGPD_OPT_NOACTION) {
 		if (conf.opts & BGPD_OPT_VERBOSE)
@@ -163,13 +173,17 @@ main(int argc, char *argv[])
 			fprintf(stderr, "configuration OK\n");
 		exit(0);
 	}
+	cflags = conf.flags;
+	connectset = &conf.connectset;
+	staticset = &conf.staticset;
+	connectset6 = &conf.connectset6;
+	staticset6 = &conf.staticset6;
 
 	if (geteuid())
 		errx(1, "need root privileges");
 
 	if (getpwnam(BGPD_USER) == NULL)
 		errx(1, "unknown user %s", BGPD_USER);
-	endpwent();
 
 	log_init(debug);
 
@@ -226,6 +240,7 @@ main(int argc, char *argv[])
 
 	while ((net = TAILQ_FIRST(&net_l)) != NULL) {
 		TAILQ_REMOVE(&net_l, net, entry);
+		filterset_free(&net->net.attrset);
 		free(net);
 	}
 
@@ -233,16 +248,15 @@ main(int argc, char *argv[])
 		TAILQ_REMOVE(rules_l, r, entry);
 		free(r);
 	}
-
-	while ((la = TAILQ_FIRST(conf.listen_addrs)) != NULL) {
-		TAILQ_REMOVE(conf.listen_addrs, la, entry);
+	TAILQ_FOREACH(la, conf.listen_addrs, entry) {
 		close(la->fd);
-		free(la);
+		la->fd = -1;
 	}
 
 	mrt_reconfigure(&mrt_l);
 
 	while (quit == 0) {
+		bzero(pfd, sizeof(pfd));
 		pfd[PFD_PIPE_SESSION].fd = ibuf_se->fd;
 		pfd[PFD_PIPE_SESSION].events = POLLIN;
 		if (ibuf_se->w.queued)
@@ -264,32 +278,29 @@ main(int argc, char *argv[])
 				quit = 1;
 			}
 
-		if (nfds > 0 && (pfd[PFD_PIPE_SESSION].revents & POLLOUT))
+		if (nfds > 0 && pfd[PFD_PIPE_SESSION].revents & POLLOUT)
 			if (msgbuf_write(&ibuf_se->w) < 0) {
 				log_warn("pipe write error (to SE)");
 				quit = 1;
 			}
 
-		if (nfds > 0 && (pfd[PFD_PIPE_ROUTE].revents & POLLOUT))
+		if (nfds > 0 && pfd[PFD_PIPE_ROUTE].revents & POLLOUT)
 			if (msgbuf_write(&ibuf_rde->w) < 0) {
 				log_warn("pipe write error (to RDE)");
 				quit = 1;
 			}
 
 		if (nfds > 0 && pfd[PFD_PIPE_SESSION].revents & POLLIN) {
-			nfds--;
 			if (dispatch_imsg(ibuf_se, PFD_PIPE_SESSION) == -1)
 				quit = 1;
 		}
 
 		if (nfds > 0 && pfd[PFD_PIPE_ROUTE].revents & POLLIN) {
-			nfds--;
 			if (dispatch_imsg(ibuf_rde, PFD_PIPE_ROUTE) == -1)
 				quit = 1;
 		}
 
 		if (nfds > 0 && pfd[PFD_SOCK_ROUTE].revents & POLLIN) {
-			nfds--;
 			if (kr_dispatch_msg() == -1)
 				quit = 1;
 		}
@@ -312,7 +323,7 @@ main(int argc, char *argv[])
 			}
 		}
 
-		if (mrtdump == 1) {
+		if (mrtdump) {
 			mrtdump = 0;
 			mrt_handler(&mrt_l);
 		}
@@ -334,9 +345,15 @@ main(int argc, char *argv[])
 		LIST_REMOVE(m, entry);
 		free(m);
 	}
+	while ((la = TAILQ_FIRST(conf.listen_addrs)) != NULL) {
+		TAILQ_REMOVE(conf.listen_addrs, la, entry);
+		close(la->fd);
+		free(la);
+	}
 
 	free(rules_l);
-	control_cleanup();
+	control_cleanup(SOCKET_NAME);
+	control_cleanup(conf.rcsock);
 	kr_shutdown();
 	pftable_clear_all();
 	free(conf.listen_addrs);
@@ -377,21 +394,14 @@ check_child(pid_t pid, const char *pname)
 }
 
 int
-send_filterset(struct imsgbuf *i, struct filter_set_head *set, int id, int f)
+send_filterset(struct imsgbuf *i, struct filter_set_head *set, int id)
 {
 	struct filter_set	*s;
 
-	for (s = SIMPLEQ_FIRST(set); s != NULL; ) {
+	TAILQ_FOREACH(s, set, entry)
 		if (imsg_compose(i, IMSG_FILTER_SET, id, 0, -1, s,
 		    sizeof(struct filter_set)) == -1)
 			return (-1);
-		if (f) {
-			SIMPLEQ_REMOVE_HEAD(set, entry);
-			free(s);
-			s = SIMPLEQ_FIRST(set);
-		} else
-			s = SIMPLEQ_NEXT(s, entry);
-	}
 	return (0);
 }
 
@@ -411,53 +421,67 @@ reconfigure(char *conffile, struct bgpd_config *conf, struct mrt_head *mrt_l,
 		return (-1);
 	}
 
+	cflags = conf->flags;
+	connectset = &conf->connectset;
+	staticset = &conf->staticset;
+	connectset6 = &conf->connectset6;
+	staticset6 = &conf->staticset6;
+
 	prepare_listeners(conf);
 
+	/* start reconfiguration */
 	if (imsg_compose(ibuf_se, IMSG_RECONF_CONF, 0, 0, -1,
 	    conf, sizeof(struct bgpd_config)) == -1)
 		return (-1);
 	if (imsg_compose(ibuf_rde, IMSG_RECONF_CONF, 0, 0, -1,
 	    conf, sizeof(struct bgpd_config)) == -1)
 		return (-1);
-	for (p = *peer_l; p != NULL; p = p->next) {
+
+	/* send peer list and listeners to the SE */
+	for (p = *peer_l; p != NULL; p = p->next)
 		if (imsg_compose(ibuf_se, IMSG_RECONF_PEER, p->conf.id, 0, -1,
 		    &p->conf, sizeof(struct peer_config)) == -1)
 			return (-1);
-		if (send_filterset(ibuf_se, &p->conf.attrset,
-		    p->conf.id, 0) == -1)
+
+	TAILQ_FOREACH(la, conf->listen_addrs, entry) {
+		if (imsg_compose(ibuf_se, IMSG_RECONF_LISTENER, 0, 0, la->fd,
+		    la, sizeof(struct listen_addr)) == -1)
 			return (-1);
+		la->fd = -1;
 	}
+
+	/* networks for the RDE */
 	while ((n = TAILQ_FIRST(&net_l)) != NULL) {
 		if (imsg_compose(ibuf_rde, IMSG_NETWORK_ADD, 0, 0, -1,
 		    &n->net, sizeof(struct network_config)) == -1)
 			return (-1);
-		if (send_filterset(ibuf_rde, &n->net.attrset, 0, 1) == -1)
+		if (send_filterset(ibuf_rde, &n->net.attrset, 0) == -1)
 			return (-1);
 		if (imsg_compose(ibuf_rde, IMSG_NETWORK_DONE, 0, 0, -1,
 		    NULL, 0) == -1)
 			return (-1);
 		TAILQ_REMOVE(&net_l, n, entry);
+		filterset_free(&n->net.attrset);
 		free(n);
 	}
+
+	/* redistribute list needs to be reloaded too */
+	if (kr_redist_reload() == -1)
+		return (-1);
+
+	/* filters for the RDE */
 	while ((r = TAILQ_FIRST(rules_l)) != NULL) {
 		if (imsg_compose(ibuf_rde, IMSG_RECONF_FILTER, 0, 0, -1,
 		    r, sizeof(struct filter_rule)) == -1)
 			return (-1);
-		if (send_filterset(ibuf_rde, &r->set, 0, 1) == -1)
+		if (send_filterset(ibuf_rde, &r->set, 0) == -1)
 			return (-1);
 		TAILQ_REMOVE(rules_l, r, entry);
+		filterset_free(&r->set);
 		free(r);
 	}
-	while ((la = TAILQ_FIRST(conf->listen_addrs)) != NULL) {
-		if (imsg_compose(ibuf_se, IMSG_RECONF_LISTENER, 0, 0, la->fd,
-		    la, sizeof(struct listen_addr)) == -1)
-			return (-1);
-		TAILQ_REMOVE(conf->listen_addrs, la, entry);
-		free(la);
-	}
-	free(conf->listen_addrs);
-	conf->listen_addrs = NULL;
 
+	/* singal both childs to replace their config */
 	if (imsg_compose(ibuf_se, IMSG_RECONF_DONE, 0, 0, -1, NULL, 0) == -1 ||
 	    imsg_compose(ibuf_rde, IMSG_RECONF_DONE, 0, 0, -1, NULL, 0) == -1)
 		return (-1);
@@ -472,6 +496,7 @@ dispatch_imsg(struct imsgbuf *ibuf, int idx)
 {
 	struct imsg		 imsg;
 	int			 n;
+	int			 rv;
 
 	if ((n = imsg_read(ibuf)) == -1)
 		return (-1);
@@ -481,6 +506,7 @@ dispatch_imsg(struct imsgbuf *ibuf, int idx)
 		return (-1);
 	}
 
+	rv = 0;
 	for (;;) {
 		if ((n = imsg_get(ibuf, &imsg)) == -1)
 			return (-1);
@@ -493,13 +519,25 @@ dispatch_imsg(struct imsgbuf *ibuf, int idx)
 			if (idx != PFD_PIPE_ROUTE)
 				log_warnx("route request not from RDE");
 			else if (kr_change(imsg.data))
-				return (-1);
+				rv = -1;
 			break;
 		case IMSG_KROUTE_DELETE:
 			if (idx != PFD_PIPE_ROUTE)
 				log_warnx("route request not from RDE");
 			else if (kr_delete(imsg.data))
-				return (-1);
+				rv = -1;
+			break;
+		case IMSG_KROUTE6_CHANGE:
+			if (idx != PFD_PIPE_ROUTE)
+				log_warnx("route request not from RDE");
+			else if (kr6_change(imsg.data))
+				rv = -1;
+			break;
+		case IMSG_KROUTE6_DELETE:
+			if (idx != PFD_PIPE_ROUTE)
+				log_warnx("route request not from RDE");
+			else if (kr6_delete(imsg.data))
+				rv = -1;
 			break;
 		case IMSG_NEXTHOP_ADD:
 			if (idx != PFD_PIPE_ROUTE)
@@ -509,7 +547,7 @@ dispatch_imsg(struct imsgbuf *ibuf, int idx)
 				    sizeof(struct bgpd_addr))
 					log_warnx("wrong imsg len");
 				else if (kr_nexthop_add(imsg.data) == -1)
-					return (-1);
+					rv = -1;
 			break;
 		case IMSG_NEXTHOP_REMOVE:
 			if (idx != PFD_PIPE_ROUTE)
@@ -529,7 +567,7 @@ dispatch_imsg(struct imsgbuf *ibuf, int idx)
 				    sizeof(struct pftable_msg))
 					log_warnx("wrong imsg len");
 				else if (pftable_addr_add(imsg.data) != 0)
-					return (-1);
+					rv = -1;
 			break;
 		case IMSG_PFTABLE_REMOVE:
 			if (idx != PFD_PIPE_ROUTE)
@@ -539,7 +577,7 @@ dispatch_imsg(struct imsgbuf *ibuf, int idx)
 				    sizeof(struct pftable_msg))
 					log_warnx("wrong imsg len");
 				else if (pftable_addr_remove(imsg.data) != 0)
-					return (-1);
+					rv = -1;
 			break;
 		case IMSG_PFTABLE_COMMIT:
 			if (idx != PFD_PIPE_ROUTE)
@@ -548,7 +586,7 @@ dispatch_imsg(struct imsgbuf *ibuf, int idx)
 				if (imsg.hdr.len != IMSG_HEADER_SIZE)
 					log_warnx("wrong imsg len");
 				else if (pftable_commit() != 0)
-					return (-1);
+					rv = -1;
 			break;
 		case IMSG_CTL_RELOAD:
 			if (idx != PFD_PIPE_SESSION)
@@ -589,6 +627,8 @@ dispatch_imsg(struct imsgbuf *ibuf, int idx)
 			break;
 		}
 		imsg_free(&imsg);
+		if (rv != 0)
+			return (rv);
 	}
 	return (0);
 }
@@ -622,3 +662,57 @@ send_imsg_session(int type, pid_t pid, void *data, u_int16_t datalen)
 {
 	imsg_compose(ibuf_se, type, 0, pid, -1, data, datalen);
 }
+
+int
+bgpd_redistribute(int type, struct kroute *kr, struct kroute6 *kr6)
+{
+	struct network_config	 net;
+	struct filter_set_head	*h;
+
+	if ((cflags & BGPD_FLAG_REDIST_CONNECTED) && kr &&
+	    (kr->flags & F_CONNECTED))
+		h = connectset;
+	else if ((cflags & BGPD_FLAG_REDIST_STATIC) && kr &&
+	    (kr->flags & F_STATIC))
+		h = staticset;
+	else if ((cflags & BGPD_FLAG_REDIST6_CONNECTED) && kr6 &&
+	    (kr6->flags & F_CONNECTED))
+		h = connectset6;
+	else if ((cflags & BGPD_FLAG_REDIST6_STATIC) && kr6 &&
+	    (kr6->flags & F_STATIC))
+		h = staticset6;
+	else
+		return (0);
+
+	bzero(&net, sizeof(net));
+	if (kr && kr6)
+		fatalx("bgpd_redistribute: unable to redistribute v4 and v6"
+		    "together");
+	if (kr != NULL) {
+		net.prefix.af = AF_INET;
+		net.prefix.v4.s_addr = kr->prefix.s_addr;
+		net.prefixlen = kr->prefixlen;
+	}
+	if (kr6 != NULL) {
+		net.prefix.af = AF_INET6;
+		memcpy(&net.prefix.v6, &kr6->prefix, sizeof(struct in6_addr));
+		net.prefixlen = kr6->prefixlen;
+	}
+
+
+	if (imsg_compose(ibuf_rde, type, 0, 0, -1, &net,
+	    sizeof(struct network_config)) == -1)
+		return (-1);
+
+	/* networks that get deleted don't need to send the filter set */
+	if (type == IMSG_NETWORK_REMOVE)
+		return (1);
+
+	if (send_filterset(ibuf_rde, h, 0) == -1)
+		return (-1);
+	if (imsg_compose(ibuf_rde, IMSG_NETWORK_DONE, 0, 0, -1, NULL, 0) == -1)
+		return (-1);
+
+	return (1);
+}
+
