@@ -1,4 +1,4 @@
-/*	$OpenBSD: bgpd.h,v 1.159 2005/03/15 10:18:39 henning Exp $ */
+/*	$OpenBSD: bgpd.h,v 1.192 2006/02/10 14:34:40 claudio Exp $ */
 
 /*
  * Copyright (c) 2003, 2004 Henning Brauer <henning@openbsd.org>
@@ -21,6 +21,7 @@
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <sys/queue.h>
+#include <net/route.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <net/if.h>
@@ -52,9 +53,14 @@
 #define	BGPD_FLAG_NO_FIB_UPDATE		0x0001
 #define	BGPD_FLAG_NO_EVALUATE		0x0002
 #define	BGPD_FLAG_REFLECTOR		0x0004
+#define	BGPD_FLAG_REDIST_STATIC		0x0008
+#define	BGPD_FLAG_REDIST_CONNECTED	0x0010
+#define	BGPD_FLAG_REDIST6_STATIC	0x0020
+#define	BGPD_FLAG_REDIST6_CONNECTED	0x0040
 #define	BGPD_FLAG_DECISION_MASK		0x0f00
 #define	BGPD_FLAG_DECISION_ROUTEAGE	0x0100
 #define	BGPD_FLAG_DECISION_TRANS_AS	0x0200
+#define	BGPD_FLAG_DECISION_MED_ALWAYS	0x0400
 
 #define	BGPD_LOG_UPDATES		0x0001
 
@@ -66,9 +72,10 @@
 #define	F_NEXTHOP		0x0008
 #define	F_DOWN			0x0010
 #define	F_STATIC		0x0020
-#define	F_LONGER		0x0040
+#define	F_DYNAMIC		0x0040
 #define	F_REJECT		0x0080
 #define	F_BLACKHOLE		0x0100
+#define	F_LONGER		0x0200
 
 enum {
 	PROC_MAIN,
@@ -93,7 +100,7 @@ struct buf {
 };
 
 struct msgbuf {
-	TAILQ_HEAD(bufs, buf)	 bufs;
+	TAILQ_HEAD(, buf)	 bufs;
 	u_int32_t		 queued;
 	int			 fd;
 };
@@ -127,9 +134,15 @@ struct listen_addr {
 };
 
 TAILQ_HEAD(listen_addrs, listen_addr);
+TAILQ_HEAD(filter_set_head, filter_set);
 
 struct bgpd_config {
+	struct filter_set_head			 connectset;
+	struct filter_set_head			 connectset6;
+	struct filter_set_head			 staticset;
+	struct filter_set_head			 staticset6;
 	struct listen_addrs			*listen_addrs;
+	char					*rcsock;
 	int					 opts;
 	int					 flags;
 	int					 log;
@@ -143,7 +156,7 @@ struct bgpd_config {
 struct buf_read {
 	u_char			 buf[READ_BUF_SIZE];
 	u_char			*rptr;
-	ssize_t			 wpos;
+	size_t			 wpos;
 };
 
 enum announce_type {
@@ -195,14 +208,11 @@ struct capabilities {
 	u_int8_t	refresh;	/* route refresh, RFC 2918 */
 };
 
-SIMPLEQ_HEAD(filter_set_head, filter_set);
-
 struct peer_config {
 	struct bgpd_addr	 remote_addr;
 	struct bgpd_addr	 local_addr;
 	struct peer_auth	 auth;
 	struct capabilities	 capabilities;
-	struct filter_set_head	 attrset;
 	char			 group[PEER_DESCR_LEN];
 	char			 descr[PEER_DESCR_LEN];
 	char			 if_depend[IFNAMSIZ];
@@ -221,8 +231,11 @@ struct peer_config {
 	u_int8_t		 ebgp;		/* 1 = ebgp, 0 = ibgp */
 	u_int8_t		 distance;	/* 1 = direct, >1 = multihop */
 	u_int8_t		 passive;
+	u_int8_t		 down;
 	u_int8_t		 announce_capa;
 	u_int8_t		 reflector_client;
+	u_int8_t		 softreconfig_in;
+	u_int8_t		 softreconfig_out;
 };
 
 struct network_config {
@@ -273,6 +286,8 @@ enum imsg_type {
 	IMSG_MRT_CLOSE,
 	IMSG_KROUTE_CHANGE,
 	IMSG_KROUTE_DELETE,
+	IMSG_KROUTE6_CHANGE,
+	IMSG_KROUTE6_DELETE,
 	IMSG_NEXTHOP_ADD,
 	IMSG_NEXTHOP_REMOVE,
 	IMSG_NEXTHOP_UPDATE,
@@ -284,7 +299,6 @@ enum imsg_type {
 	IMSG_NETWORK_FLUSH,
 	IMSG_NETWORK_DONE,
 	IMSG_FILTER_SET,
-	IMSG_CTL_SHOW_NEIGHBOR,
 	IMSG_CTL_END,
 	IMSG_CTL_RELOAD,
 	IMSG_CTL_FIB_COUPLE,
@@ -293,13 +307,19 @@ enum imsg_type {
 	IMSG_CTL_NEIGHBOR_DOWN,
 	IMSG_CTL_NEIGHBOR_CLEAR,
 	IMSG_CTL_KROUTE,
+	IMSG_CTL_KROUTE6,
 	IMSG_CTL_KROUTE_ADDR,
+	IMSG_CTL_RESULT,
+	IMSG_CTL_SHOW_NEIGHBOR,
 	IMSG_CTL_SHOW_NEXTHOP,
 	IMSG_CTL_SHOW_INTERFACE,
 	IMSG_CTL_SHOW_RIB,
 	IMSG_CTL_SHOW_RIB_AS,
 	IMSG_CTL_SHOW_RIB_PREFIX,
 	IMSG_CTL_SHOW_NETWORK,
+	IMSG_CTL_SHOW_NETWORK6,
+	IMSG_CTL_SHOW_RIB_MEM,
+	IMSG_CTL_SHOW_TERSE,
 	IMSG_REFRESH,
 	IMSG_IFINFO
 };
@@ -316,6 +336,12 @@ struct imsg {
 	void		*data;
 };
 
+enum ctl_results {
+	CTL_RES_OK,
+	CTL_RES_NOSUCHPEER,
+	CTL_RES_DENIED
+};
+
 /* needed for session.h parse prototype */
 LIST_HEAD(mrt_head, mrt);
 
@@ -359,6 +385,7 @@ struct kroute {
 	struct in_addr	prefix;
 	struct in_addr	nexthop;
 	u_int16_t	flags;
+	u_int16_t	labelid;
 	u_short		ifindex;
 	u_int8_t	prefixlen;
 };
@@ -367,6 +394,7 @@ struct kroute6 {
 	struct in6_addr	prefix;
 	struct in6_addr	nexthop;
 	u_int16_t	flags;
+	u_int16_t	labelid;
 	u_short		ifindex;
 	u_int8_t	prefixlen;
 };
@@ -417,6 +445,16 @@ struct ctl_neighbor {
 	char			descr[PEER_DESCR_LEN];
 };
 
+struct kroute_label {
+	struct kroute	kr;
+	char		label[RTLABEL_LEN];
+};
+
+struct kroute6_label {
+	struct kroute6	kr;
+	char		label[RTLABEL_LEN];
+};
+
 #define	F_RIB_ELIGIBLE	0x01
 #define	F_RIB_ACTIVE	0x02
 #define	F_RIB_INTERNAL	0x04
@@ -428,20 +466,21 @@ struct ctl_show_rib {
 	time_t			lastchange;
 	u_int32_t		local_pref;
 	u_int32_t		med;
-	u_int16_t		prefix_cnt;
-	u_int16_t		active_cnt;
+	u_int32_t		prefix_cnt;
+	u_int32_t		active_cnt;
+	u_int32_t		adjrib_cnt;
 	u_int16_t		aspath_len;
+	u_int16_t		flags;
 	u_int8_t		prefixlen;
 	u_int8_t		origin;
-	u_int8_t		flags;
 	/* plus a aspath_len bytes long aspath */
 };
 
 struct ctl_show_rib_prefix {
 	struct bgpd_addr	prefix;
 	time_t			lastchange;
+	u_int16_t		flags;
 	u_int8_t		prefixlen;
-	u_int8_t		flags;
 };
 
 enum as_spec {
@@ -465,9 +504,7 @@ enum filter_actions {
 
 enum directions {
 	DIR_IN = 1,
-	DIR_OUT,
-	DIR_DEFAULT_IN,		/* only needed to apply default set */
-	DIR_DEFAULT_OUT
+	DIR_OUT
 };
 
 enum from_spec {
@@ -497,6 +534,7 @@ struct filter_peers {
 /* special community type */
 #define	COMMUNITY_ERROR			-1
 #define	COMMUNITY_ANY			-2
+#define	COMMUNITY_NEIGHBOR_AS		-3
 #define	COMMUNITY_WELLKNOWN		0xffff
 #define	COMMUNITY_NO_EXPORT		0xff01
 #define	COMMUNITY_NO_ADVERTISE		0xff02
@@ -544,6 +582,8 @@ enum action_types {
 	ACTION_SET_RELATIVE_LOCALPREF,
 	ACTION_SET_MED,
 	ACTION_SET_RELATIVE_MED,
+	ACTION_SET_WEIGHT,
+	ACTION_SET_RELATIVE_WEIGHT,
 	ACTION_SET_PREPEND_SELF,
 	ACTION_SET_PREPEND_PEER,
 	ACTION_SET_NEXTHOP,
@@ -551,19 +591,45 @@ enum action_types {
 	ACTION_SET_NEXTHOP_BLACKHOLE,
 	ACTION_SET_NEXTHOP_NOMODIFY,
 	ACTION_SET_COMMUNITY,
-/*	ACTION_SCRUB_COMMUNITY, */
-	ACTION_PFTABLE
+	ACTION_DEL_COMMUNITY,
+	ACTION_PFTABLE,
+	ACTION_PFTABLE_ID,
+	ACTION_RTLABEL,
+	ACTION_RTLABEL_ID
+};
+
+static const char * const filterset_names[] = {
+	"localpref",
+	"localpref",
+	"metric",
+	"metric",
+	"weight",
+	"weight",
+	"prepend-self",
+	"prepend-peer",
+	"nexthop",
+	"nexthop",
+	"nexthop",
+	"nexthop",
+	"community",
+	"community delete",
+	"pftable",
+	"pftable",
+	"rtlabel",
+	"rtlabel"
 };
 
 struct filter_set {
-	SIMPLEQ_ENTRY(filter_set)	entry;
+	TAILQ_ENTRY(filter_set)		entry;
 	union {
 		u_int8_t		prepend;
+		u_int16_t		id;
 		u_int32_t		metric;
 		int32_t			relative;
 		struct bgpd_addr	nexthop;
 		struct filter_community	community;
 		char			pftable[PFTABLE_LEN];
+		char			rtlabel[RTLABEL_LEN];
 	} action;
 	enum action_types		type;
 };
@@ -573,6 +639,21 @@ struct rrefresh {
 	u_int8_t	safi;
 };
 
+struct rde_memstats {
+	int64_t		path_cnt;
+	int64_t		prefix_cnt;
+	int64_t		pt4_cnt;
+	int64_t		pt6_cnt;
+	int64_t		nexthop_cnt;
+	int64_t		aspath_cnt;
+	int64_t		aspath_size;
+	int64_t		aspath_refs;
+	int64_t		attr_cnt;
+	int64_t		attr_refs;
+	int64_t		attr_data;
+	int64_t		attr_dcnt;
+};
+
 /* Address Family Numbers as per rfc1700 */
 #define	AFI_IPv4	1
 #define	AFI_IPv6	2
@@ -589,10 +670,11 @@ struct rrefresh {
 /* bgpd.c */
 void		 send_nexthop_update(struct kroute_nexthop *);
 void		 send_imsg_session(int, pid_t, void *, u_int16_t);
+int		 bgpd_redistribute(int, struct kroute *, struct kroute6 *);
 
 /* buffer.c */
 struct buf	*buf_open(size_t);
-int		 buf_add(struct buf *, void *, size_t);
+int		 buf_add(struct buf *, const void *, size_t);
 void		*buf_reserve(struct buf *, size_t);
 int		 buf_close(struct msgbuf *, struct buf *);
 int		 buf_write(int, struct buf *);
@@ -614,8 +696,6 @@ void		 log_info(const char *, ...);
 void		 log_debug(const char *, ...);
 void		 fatal(const char *);
 void		 fatalx(const char *);
-void		 fatal_ensure(const char *, int, const char *);
-const char	*log_addr(const struct bgpd_addr *);
 
 /* parse.y */
 int	 cmdline_symset(char *);
@@ -629,18 +709,20 @@ void	 imsg_init(struct imsgbuf *, int);
 int	 imsg_read(struct imsgbuf *);
 int	 imsg_get(struct imsgbuf *, struct imsg *);
 int	 imsg_compose(struct imsgbuf *, enum imsg_type, u_int32_t, pid_t, int,
-	    void *, u_int16_t);
+	    const void *, u_int16_t);
 struct buf	*imsg_create(struct imsgbuf *, enum imsg_type, u_int32_t, pid_t,
 		    u_int16_t);
-int	 imsg_add(struct buf *, void *, u_int16_t);
+int	 imsg_add(struct buf *, const void *, u_int16_t);
 int	 imsg_close(struct imsgbuf *, struct buf *);
 void	 imsg_free(struct imsg *);
 int	 imsg_get_fd(struct imsgbuf *);
 
 /* kroute.c */
 int		 kr_init(int);
-int		 kr_change(struct kroute *);
-int		 kr_delete(struct kroute *);
+int		 kr_change(struct kroute_label *);
+int		 kr_delete(struct kroute_label *);
+int		 kr6_change(struct kroute6_label *);
+int		 kr6_delete(struct kroute6_label *);
 void		 kr_shutdown(void);
 void		 kr_fib_couple(void);
 void		 kr_fib_decouple(void);
@@ -649,14 +731,14 @@ int		 kr_nexthop_add(struct bgpd_addr *);
 void		 kr_nexthop_delete(struct bgpd_addr *);
 void		 kr_show_route(struct imsg *);
 void		 kr_ifinfo(char *);
+int		 kr_redist_reload(void);
 in_addr_t	 prefixlen2mask(u_int8_t);
 struct in6_addr	*prefixlen2mask6(u_int8_t prefixlen);
 void		 inet6applymask(struct in6_addr *, const struct in6_addr *,
 		    int);
 
 /* control.c */
-int	control_init(void);
-void	control_cleanup(void);
+void	control_cleanup(const char *);
 int	control_imsg_relay(struct imsg *);
 
 /* pftable.c */
@@ -667,4 +749,27 @@ int	pftable_addr_add(struct pftable_msg *);
 int	pftable_addr_remove(struct pftable_msg *);
 int	pftable_commit(void);
 
+/* name2id.c */
+u_int16_t	 rtlabel_name2id(const char *);
+const char	*rtlabel_id2name(u_int16_t);
+void		 rtlabel_unref(u_int16_t);
+void		 rtlabel_ref(u_int16_t);
+u_int16_t	 pftable_name2id(const char *);
+const char	*pftable_id2name(u_int16_t);
+void		 pftable_unref(u_int16_t);
+void		 pftable_ref(u_int16_t);
+
+
+/* rde_filter.c */
+void		 filterset_free(struct filter_set_head *);
+int		 filterset_cmp(struct filter_set *, struct filter_set *);
+
+/* util.c */
+const char	*log_addr(const struct bgpd_addr *);
+const char	*log_in6addr(const struct in6_addr *);
+const char *	 log_sockaddr(struct sockaddr *);
+int		 aspath_snprint(char *, size_t, void *, u_int16_t);
+int		 aspath_asprint(char **, void *, u_int16_t);
+size_t		 aspath_strlen(void *, u_int16_t);
+
 #endif /* __BGPD_H__ */
