@@ -1,4 +1,4 @@
-/*	$OpenBSD: imsg.c,v 1.34 2005/03/09 14:08:03 henning Exp $ */
+/*	$OpenBSD: imsg.c,v 1.37 2005/11/02 15:34:43 claudio Exp $ */
 
 /*
  * Copyright (c) 2003, 2004 Henning Brauer <henning@openbsd.org>
@@ -86,8 +86,7 @@ imsg_read(struct imsgbuf *ibuf)
 int
 imsg_get(struct imsgbuf *ibuf, struct imsg *imsg)
 {
-	ssize_t			 datalen = 0;
-	size_t			 av, left;
+	size_t			 av, left, datalen;
 
 	av = ibuf->r.wpos;
 
@@ -123,7 +122,7 @@ imsg_get(struct imsgbuf *ibuf, struct imsg *imsg)
 
 int
 imsg_compose(struct imsgbuf *ibuf, enum imsg_type type, u_int32_t peerid,
-    pid_t pid, int fd, void *data, u_int16_t datalen)
+    pid_t pid, int fd, const void *data, u_int16_t datalen)
 {
 	struct buf	*wbuf;
 	int		 n;
@@ -137,7 +136,7 @@ imsg_compose(struct imsgbuf *ibuf, enum imsg_type type, u_int32_t peerid,
 	wbuf->fd = fd;
 
 	if ((n = imsg_close(ibuf, wbuf)) < 0)
-			return (-1);
+		return (-1);
 
 	return (n);
 }
@@ -150,7 +149,7 @@ imsg_create(struct imsgbuf *ibuf, enum imsg_type type, u_int32_t peerid,
 	struct imsg_hdr	 hdr;
 
 	if (datalen > MAX_IMSGSIZE - IMSG_HEADER_SIZE) {
-		log_warnx("imsg_create_core: len %u > MAX_IMSGSIZE; "
+		log_warnx("imsg_create: len %u > MAX_IMSGSIZE; "
 		    "type %u peerid %lu", datalen + IMSG_HEADER_SIZE,
 		    type, peerid);
 		return (NULL);
@@ -171,7 +170,7 @@ imsg_create(struct imsgbuf *ibuf, enum imsg_type type, u_int32_t peerid,
 }
 
 int
-imsg_add(struct buf *msg, void *data, u_int16_t datalen)
+imsg_add(struct buf *msg, const void *data, u_int16_t datalen)
 {
 	if (datalen)
 		if (buf_add(msg, data, datalen) == -1) {
