@@ -1,4 +1,4 @@
-/*	$OpenBSD: printconf.c,v 1.37 2005/03/14 17:32:04 claudio Exp $	*/
+/*	$OpenBSD: printconf.c,v 1.53 2006/02/10 14:34:40 claudio Exp $	*/
 
 /*
  * Copyright (c) 2003, 2004 Henning Brauer <henning@openbsd.org>
@@ -32,6 +32,7 @@ void		 print_peer(struct peer_config *, struct bgpd_config *,
 		    const char *);
 const char	*print_auth_alg(u_int8_t);
 const char	*print_enc_alg(u_int8_t);
+const char	*print_safi(u_int8_t);
 void		 print_rule(struct peer *, struct filter_rule *);
 const char *	 mrt_type(enum mrt_type);
 void		 print_mrt(u_int32_t, u_int32_t, const char *, const char *);
@@ -77,11 +78,11 @@ print_set(struct filter_set_head *set)
 {
 	struct filter_set	*s;
 
-	if (SIMPLEQ_EMPTY(set))
+	if (TAILQ_EMPTY(set))
 		return;
 
 	printf("set { ");
-	SIMPLEQ_FOREACH(s, set, entry) {
+	TAILQ_FOREACH(s, set, entry) {
 		switch (s->type) {
 		case ACTION_SET_LOCALPREF:
 			printf("localpref %u ", s->action.metric);
@@ -95,6 +96,12 @@ print_set(struct filter_set_head *set)
 		case ACTION_SET_RELATIVE_MED:
 			printf("metric %+d ", s->action.relative);
 			break;
+		case ACTION_SET_WEIGHT:
+			printf("weight %u ", s->action.metric);
+			break;
+		case ACTION_SET_RELATIVE_WEIGHT:
+			printf("weight %+d ", s->action.relative);
+			break;
 		case ACTION_SET_NEXTHOP:
 			printf("nexthop %s ", log_addr(&s->action.nexthop));
 			break;
@@ -113,6 +120,10 @@ print_set(struct filter_set_head *set)
 		case ACTION_SET_PREPEND_PEER:
 			printf("prepend-neighbor %u ", s->action.prepend);
 			break;
+		case ACTION_DEL_COMMUNITY:
+			printf("community delete %u:%u ",
+			    s->action.community.as, s->action.community.type);
+			break;
 		case ACTION_SET_COMMUNITY:
 			printf("community %u:%u ", s->action.community.as,
 			    s->action.community.type);
@@ -120,6 +131,14 @@ print_set(struct filter_set_head *set)
 		case ACTION_PFTABLE:
 			printf("pftable %s ", s->action.pftable);
 			break;
+		case ACTION_RTLABEL:
+			printf("rtlabel %s ", s->action.rtlabel);
+			break;
+		case ACTION_RTLABEL_ID:
+		case ACTION_PFTABLE_ID:
+			/* not possible */
+			printf("king bula saiz: config broken");
+			break;
 		}
 	}
 	printf("}");
@@ -150,6 +169,9 @@ print_mainconf(struct bgpd_config *conf)
 	if (conf->flags & BGPD_FLAG_DECISION_ROUTEAGE)
 		printf("rde route-age evaluate\n");
 
+	if (conf->flags & BGPD_FLAG_DECISION_MED_ALWAYS)
+		printf("rde med compare always\n");
+
 	if (conf->flags & BGPD_FLAG_DECISION_TRANS_AS)
 		printf("transparent-as yes\n");
 
@@ -159,13 +181,42 @@ print_mainconf(struct bgpd_config *conf)
 	TAILQ_FOREACH(la, conf->listen_addrs, entry)
 		printf("listen on %s\n",
 		    log_sockaddr((struct sockaddr *)&la->sa));
+
+	if (conf->flags & BGPD_FLAG_REDIST_CONNECTED) {
+		printf("network inet connected");
+		if (!TAILQ_EMPTY(&conf->connectset))
+			printf(" ");
+		print_set(&conf->connectset);
+		printf("\n");
+	}
+	if (conf->flags & BGPD_FLAG_REDIST_STATIC) {
+		printf("network inet static");
+		if (!TAILQ_EMPTY(&conf->staticset))
+			printf(" ");
+		print_set(&conf->staticset);
+		printf("\n");
+	}
+	if (conf->flags & BGPD_FLAG_REDIST6_CONNECTED) {
+		printf("network inet6 connected");
+		if (!TAILQ_EMPTY(&conf->connectset6))
+			printf(" ");
+		print_set(&conf->connectset6);
+		printf("\n");
+	}
+	if (conf->flags & BGPD_FLAG_REDIST_STATIC) {
+		printf("network inet6 static");
+		if (!TAILQ_EMPTY(&conf->staticset6))
+			printf(" ");
+		print_set(&conf->staticset6);
+		printf("\n");
+	}
 }
 
 void
 print_network(struct network_config *n)
 {
 	printf("network %s/%u", log_addr(&n->prefix), n->prefixlen);
-	if (!SIMPLEQ_EMPTY(&n->attrset))
+	if (!TAILQ_EMPTY(&n->attrset))
 		printf(" ");
 	print_set(&n->attrset);
 	printf("\n");
@@ -253,11 +304,19 @@ print_peer(struct peer_config *p, struct bgpd_config *conf, const char *c)
 	else if (p->auth.method == AUTH_IPSEC_IKE_ESP)
 		printf("%s\tipsec esp ike\n", c);
 
-	if (!SIMPLEQ_EMPTY(&p->attrset))
-		printf("%s\t", c);
-	print_set(&p->attrset);
-	if (!SIMPLEQ_EMPTY(&p->attrset))
-		printf("\n");
+	printf("%s\tannounce IPv4 %s\n", c, print_safi(p->capabilities.mp_v4));
+	printf("%s\tannounce IPv6 %s\n", c, print_safi(p->capabilities.mp_v6));
+
+	if (p->softreconfig_in == 1)
+		printf("%s\tsoftreconfig in yes\n", c);
+	else
+		printf("%s\tsoftreconfig in no\n", c);
+
+	if (p->softreconfig_out == 1)
+		printf("%s\tsoftreconfig out yes\n", c);
+	else
+		printf("%s\tsoftreconfig out no\n", c);
+
 
 	print_mrt(p->id, p->groupid, c, "\t");
 
@@ -290,6 +349,19 @@ print_enc_alg(u_int8_t alg)
 	}
 }
 
+const char *
+print_safi(u_int8_t safi)
+{
+	switch (safi) {
+	case SAFI_NONE:
+		return ("none");
+	case SAFI_UNICAST:
+		return ("unicast");
+	default:
+		return ("?");
+	}
+}
+
 void
 print_rule(struct peer *peer_l, struct filter_rule *r)
 {
@@ -317,7 +389,7 @@ print_rule(struct peer *peer_l, struct filter_rule *r)
 		    p = p->next)
 			;	/* nothing */
 		if (p == NULL)
-			printf("?");
+			printf("? ");
 		else
 			printf("%s ", log_addr(&p->conf.remote_addr));
 	} else if (r->peer.groupid) {
@@ -327,7 +399,7 @@ print_rule(struct peer *peer_l, struct filter_rule *r)
 		if (p == NULL)
 			printf("group ? ");
 		else
-			printf("group %s ", p->conf.group);
+			printf("group \"%s\" ", p->conf.group);
 	} else
 		printf("any ");
 
@@ -360,13 +432,18 @@ print_rule(struct peer *peer_l, struct filter_rule *r)
 	}
 
 	if (r->match.community.as != 0) {
+		printf("community ");
 		if (r->match.community.as == COMMUNITY_ANY)
 			printf("*:");
+		else if (r->match.community.as == COMMUNITY_NEIGHBOR_AS)
+			printf("neighbor-as:");
 		else
 			printf("%d:", r->match.community.as);
 
 		if (r->match.community.type == COMMUNITY_ANY)
 			printf("* ");
+		else if (r->match.community.type == COMMUNITY_NEIGHBOR_AS)
+			printf("neighbor-as ");
 		else
 			printf("%d ", r->match.community.type);
 	}
@@ -488,13 +565,14 @@ print_config(struct bgpd_config *conf, struct network_head *net_l,
 	struct network		*n;
 
 	xmrt_l = mrt_l;
-	print_mainconf(conf);
 	printf("\n");
-	print_mrt(0, 0, "", "");
+	print_mainconf(conf);
 	printf("\n");
 	TAILQ_FOREACH(n, net_l, entry)
 		print_network(&n->net);
 	printf("\n");
+	print_mrt(0, 0, "", "");
+	printf("\n");
 	print_groups(conf, peer_l);
 	printf("\n");
 	TAILQ_FOREACH(r, rules_l, entry)
