@@ -1,4 +1,4 @@
-/*	$OpenBSD: rde_update.c,v 1.35 2005/03/11 12:54:20 claudio Exp $ */
+/*	$OpenBSD: rde_update.c,v 1.49 2006/01/13 13:04:33 claudio Exp $ */
 
 /*
  * Copyright (c) 2004 Claudio Jeker <claudio@openbsd.org>
@@ -17,6 +17,7 @@
  */
 #include <sys/types.h>
 #include <sys/queue.h>
+#include <sys/hash.h>
 
 #include <stdlib.h>
 #include <string.h>
@@ -24,25 +25,31 @@
 #include "bgpd.h"
 #include "rde.h"
 
-int	up_generate_attr(struct rde_peer *, struct update_attr *,
-	    struct rde_aspath *);
+in_addr_t	up_get_nexthop(struct rde_peer *, struct rde_aspath *);
+int		up_generate_mp_reach(struct rde_peer *, struct update_attr *,
+		    struct rde_aspath *, sa_family_t);
+int		up_generate_attr(struct rde_peer *, struct update_attr *,
+		    struct rde_aspath *, sa_family_t);
+int		up_set_prefix(u_char *, int, struct bgpd_addr *, u_int8_t);
 
 /* update stuff. */
 struct update_prefix {
-	struct bgpd_addr		 prefix;
-	int				 prefixlen;
-	struct uplist_prefix		*prefix_h;
 	TAILQ_ENTRY(update_prefix)	 prefix_l;
 	RB_ENTRY(update_prefix)		 entry;
+	struct uplist_prefix		*prefix_h;
+	struct bgpd_addr		 prefix;
+	int				 prefixlen;
 };
 
 struct update_attr {
-	u_int32_t			 attr_hash;
-	u_char				*attr;
-	u_int16_t			 attr_len;
-	struct uplist_prefix		 prefix_h;
 	TAILQ_ENTRY(update_attr)	 attr_l;
 	RB_ENTRY(update_attr)		 entry;
+	struct uplist_prefix		 prefix_h;
+	u_char				*attr;
+	u_char				*mpattr;
+	u_int32_t			 attr_hash;
+	u_int16_t			 attr_len;
+	u_int16_t			 mpattr_len;
 };
 
 void	up_clear(struct uplist_attr *, struct uplist_prefix *);
@@ -84,6 +91,7 @@ up_clear(struct uplist_attr *updates, struct uplist_prefix *withdraws)
 			free(up);
 		}
 		free(ua->attr);
+		free(ua->mpattr);
 		free(ua);
 	}
 
@@ -146,14 +154,16 @@ up_prefix_cmp(struct update_prefix *a, struct update_prefix *b)
 int
 up_attr_cmp(struct update_attr *a, struct update_attr *b)
 {
-	if (a->attr_hash < b->attr_hash)
-		return (-1);
-	if (a->attr_hash > b->attr_hash)
-		return (1);
-	if (a->attr_len < b->attr_len)
-		return (-1);
-	if (a->attr_len > b->attr_len)
-		return (1);
+	int	r;
+
+	if ((r = a->attr_hash - b->attr_hash) != 0)
+		return (r);
+	if ((r = a->attr_len - b->attr_len) != 0)
+		return (r);
+	if ((r = a->mpattr_len - b->mpattr_len) != 0)
+		return (r);
+	if ((r = memcmp(a->mpattr, b->mpattr, a->mpattr_len)) != 0)
+		return (r);
 	return (memcmp(a->attr, b->attr, a->attr_len));
 }
 
@@ -187,6 +197,7 @@ up_add(struct rde_peer *peer, struct update_prefix *p, struct update_attr *a)
 			log_warnx("uptree_attr insert failed");
 			/* cleanup */
 			free(a->attr);
+			free(a->mpattr);
 			free(a);
 			free(p);
 			return (-1);
@@ -197,6 +208,7 @@ up_add(struct rde_peer *peer, struct update_prefix *p, struct update_attr *a)
 		/* 1.2 if found -> use that, free a */
 		if (a != NULL) {
 			free(a->attr);
+			free(a->mpattr);
 			free(a);
 			a = na;
 			/* move to end of update queue */
@@ -242,336 +254,377 @@ up_add(struct rde_peer *peer, struct update_prefix *p, struct update_attr *a)
 	return (0);
 }
 
+int
+up_test_update(struct rde_peer *peer, struct prefix *p)
+{
+	struct bgpd_addr	 addr;
+	struct attr		*attr;
+
+	if (peer->state != PEER_UP)
+		return (-1);
+
+	if (p == NULL)
+		/* no prefix available */
+		return (0);
+
+	if (peer == p->aspath->peer)
+		/* Do not send routes back to sender */
+		return (0);
+
+	pt_getaddr(p->prefix, &addr);
+	switch (addr.af) {
+	case AF_INET:
+		if (peer->capa_announced.mp_v4 == SAFI_NONE &&
+		    peer->capa_received.mp_v6 != SAFI_NONE)
+			return (-1);
+		break;
+	case AF_INET6:
+		if (peer->capa_announced.mp_v6 == SAFI_NONE)
+			return (-1);
+		break;
+	}
+
+	if (peer->conf.ebgp && !aspath_loopfree(p->aspath->aspath,
+	    peer->conf.remote_as)) {
+		/*
+		 * Do not send routes back to sender which would
+		 * cause an aspath loop.
+		 */
+		return (0);
+	}
+
+	if (p->aspath->peer->conf.ebgp == 0 && peer->conf.ebgp == 0) {
+		/*
+		 * route reflector redistribution rules:
+		 * 1. if announce is set                -> announce
+		 * 2. old non-client, new non-client    -> no
+		 * 3. old client, new non-client        -> yes
+		 * 4. old non-client, new client        -> yes
+		 * 5. old client, new client            -> yes
+		 */
+		if (p->aspath->peer->conf.reflector_client == 0 &&
+		    peer->conf.reflector_client == 0 &&
+		    (p->aspath->flags & F_PREFIX_ANNOUNCED) == 0)
+			/* Do not redistribute updates to ibgp peers */
+			return (0);
+	}
+
+	/* announce type handling */
+	switch (peer->conf.announce_type) {
+	case ANNOUNCE_UNDEF:
+	case ANNOUNCE_NONE:
+	case ANNOUNCE_DEFAULT_ROUTE:
+		/*
+		 * no need to withdraw old prefix as this will be
+		 * filtered out as well.
+		 */
+		return (-1);
+	case ANNOUNCE_ALL:
+		break;
+	case ANNOUNCE_SELF:
+		/*
+		 * pass only prefix that have a aspath count
+		 * of zero this is equal to the ^$ regex.
+		 */
+		if (p->aspath->aspath->ascnt != 0)
+			return (0);
+		break;
+	}
+
+	/* well known communities */
+	if (rde_filter_community(p->aspath,
+	    COMMUNITY_WELLKNOWN, COMMUNITY_NO_ADVERTISE))
+		return (0);
+	if (peer->conf.ebgp && rde_filter_community(p->aspath,
+	    COMMUNITY_WELLKNOWN, COMMUNITY_NO_EXPORT))
+		return (0);
+	if (peer->conf.ebgp && rde_filter_community(p->aspath,
+	    COMMUNITY_WELLKNOWN, COMMUNITY_NO_EXPSUBCONFED))
+		return (0);
+
+	/*
+	 * Don't send messages back to originator
+	 * this is not specified in the RFC but seems logical.
+	 */
+	if ((attr = attr_optget(p->aspath, ATTR_ORIGINATOR_ID)) != NULL) {
+		if (memcmp(attr->data, &peer->remote_bgpid,
+		    sizeof(peer->remote_bgpid)) == 0) {
+			/* would cause loop don't send */
+			return (-1);
+		}
+	}
+
+	return (1);
+}
+
+int
+up_generate(struct rde_peer *peer, struct rde_aspath *asp,
+    struct bgpd_addr *addr, u_int8_t prefixlen)
+{
+	struct update_attr		*ua = NULL;
+	struct update_prefix		*up;
+
+	if (asp) {
+		ua = calloc(1, sizeof(struct update_attr));
+		if (ua == NULL)
+			fatal("up_generate_updates");
+
+		if (up_generate_attr(peer, ua, asp, addr->af) == -1) {
+			log_warnx("generation of bgp path attributes failed");
+			free(ua);
+			return (-1);
+		}
+		/*
+		 * use aspath_hash as attr_hash, this may be unoptimal
+		 * but currently I don't care.
+		 */
+		ua->attr_hash = hash32_buf(ua->attr, ua->attr_len, HASHINIT);
+		if (ua->mpattr)
+			ua->attr_hash = hash32_buf(ua->mpattr, ua->mpattr_len,
+			    ua->attr_hash);
+	}
+
+	up = calloc(1, sizeof(struct update_prefix));
+	if (up == NULL)
+		fatal("up_generate_updates");
+	up->prefix = *addr;
+	up->prefixlen = prefixlen;
+
+	if (up_add(peer, up, ua) == -1)
+		return (-1);
+
+	return (0);
+}
+
 void
-up_generate_updates(struct rde_peer *peer,
+up_generate_updates(struct filter_head *rules, struct rde_peer *peer,
     struct prefix *new, struct prefix *old)
 {
-	struct update_attr		*a;
-	struct update_prefix		*p;
-	struct attr			*attr;
-	struct rde_aspath		*fasp;
+	struct rde_aspath		*asp;
 	struct bgpd_addr		 addr;
 
 	if (peer->state != PEER_UP)
 		return;
 
-	if (new == NULL || (new->aspath->nexthop != NULL &&
-	    new->aspath->nexthop->state != NEXTHOP_REACH)) {
-		if (old == NULL)
-			/* new prefix got filtered and no old prefix avail */
+	if (new == NULL) {
+		if (up_test_update(peer, old) != 1)
 			return;
 
-		if (peer == old->peer)
-			/* Do not send routes back to sender */
-			return;
-
-		if (peer->conf.ebgp &&
-		    !aspath_loopfree(old->aspath->aspath,
-		    peer->conf.remote_as))
-			/*
-			 * Do not send routes back to sender which would
-			 * cause a aspath loop.
-			 */
-			return;
-
-		if (old->peer->conf.ebgp == 0 && peer->conf.ebgp == 0) {
-			/*
-			 * route reflector redistribution rules:
-			 * 1. if announce is set		-> announce
-			 * 2. old non-client, new non-client	-> no
-			 * 3. old client, new non-client	-> yes
-			 * 4. old non-client, new client	-> yes
-			 * 5. old client, new client		-> yes
-			 */
-			if (old->peer->conf.reflector_client == 0 &&
-			    peer->conf.reflector_client == 0 &&
-			    (old->aspath->flags & F_PREFIX_ANNOUNCED) == 0)
-				/* Do not redistribute updates to ibgp peers */
-				return;
-		}
-
-		/* announce type handling */
-		switch (peer->conf.announce_type) {
-		case ANNOUNCE_UNDEF:
-		case ANNOUNCE_NONE:
-		case ANNOUNCE_DEFAULT_ROUTE:
-			return;
-		case ANNOUNCE_ALL:
-			break;
-		case ANNOUNCE_SELF:
-			/*
-			 * pass only prefix that have a aspath count
-			 * of zero this is equal to the ^$ regex.
-			 */
-			if (old->aspath->aspath->ascnt != 0)
-				return;
-			break;
-		}
-
-		/* well known communities */
-		if (rde_filter_community(old->aspath,
-		    COMMUNITY_WELLKNOWN, COMMUNITY_NO_ADVERTISE))
-			return;
-		if (peer->conf.ebgp && rde_filter_community(old->aspath,
-		    COMMUNITY_WELLKNOWN, COMMUNITY_NO_EXPORT))
-			return;
-		if (peer->conf.ebgp && rde_filter_community(old->aspath,
-		    COMMUNITY_WELLKNOWN, COMMUNITY_NO_EXPSUBCONFED))
-			return;
-
-		/*
-		 * Don't send messages back to originator
-		 * this is not specified in the RFC but seems logical.
-		 */
-		if ((attr = attr_optget(old->aspath,
-		    ATTR_ORIGINATOR_ID)) != NULL) {
-			if (memcmp(attr->data, &peer->remote_bgpid,
-			    sizeof(peer->remote_bgpid)) == 0)
-				/* would cause loop don't send */
-				return;
-		}
-
-		/* copy attributes for output filter */
-		fasp = path_copy(old->aspath);
-
-		/* default override not needed here as this is a withdraw */
-
 		pt_getaddr(old->prefix, &addr);
-		if (rde_filter(peer, fasp, &addr,
-		    old->prefix->prefixlen, DIR_OUT) == ACTION_DENY) {
-			path_put(fasp);
+		if (rde_filter(NULL, rules, peer, old->aspath, &addr,
+		    old->prefix->prefixlen, old->aspath->peer, DIR_OUT) ==
+		    ACTION_DENY)
 			return;
-		}
-		path_put(fasp);
 
 		/* withdraw prefix */
-		p = calloc(1, sizeof(struct update_prefix));
-		if (p == NULL)
-			fatal("up_generate_updates");
-
-		p->prefix = addr;
-		p->prefixlen = old->prefix->prefixlen;
-		if (up_add(peer, p, NULL) == -1)
-			log_warnx("queuing withdraw failed.");
+		up_generate(peer, NULL, &addr, old->prefix->prefixlen);
 	} else {
-		if (peer == new->peer) {
-			/* Do not send routes back to sender */
-			up_generate_updates(peer, NULL, old);
-			return;
-		}
-
-		if (peer->conf.ebgp &&
-		    !aspath_loopfree(new->aspath->aspath,
-		    peer->conf.remote_as)) {
-			/*
-			 * Do not send routes back to sender which would
-			 * cause a aspath loop.
-			 */
-			up_generate_updates(peer, NULL, old);
-			return;
-		}
-
-		if (new->peer->conf.ebgp == 0 && peer->conf.ebgp == 0) {
-			/*
-			 * route reflector redistribution rules:
-			 * 1. if announce is set		-> announce
-			 * 2. old non-client, new non-client	-> no
-			 * 3. old client, new non-client	-> yes
-			 * 4. old non-client, new client	-> yes
-			 * 5. old client, new client		-> yes
-			 */
-			if (new->peer->conf.reflector_client == 0 &&
-			    peer->conf.reflector_client == 0 &&
-			    (new->aspath->flags & F_PREFIX_ANNOUNCED) == 0) {
-				/* Do not redistribute updates to ibgp peers */
-				up_generate_updates(peer, NULL, old);
-				return;
-			}
-		}
-
-		/* announce type handling */
-		switch (peer->conf.announce_type) {
-		case ANNOUNCE_UNDEF:
-		case ANNOUNCE_NONE:
-		case ANNOUNCE_DEFAULT_ROUTE:
-			/*
-			 * no need to withdraw old prefix as this will be
-			 * filtered out to.
-			 */
-			return;
-		case ANNOUNCE_ALL:
+		switch (up_test_update(peer, new)) {
+		case 1:
 			break;
-		case ANNOUNCE_SELF:
-			/*
-			 * pass only prefix that have a aspath count
-			 * of zero this is equal to the ^$ regex.
-			 */
-			if (new->aspath->aspath->ascnt != 0) {
-				up_generate_updates(peer, NULL, old);
-				return;
-			}
-			break;
-		}
-
-		/* well known communities */
-		if (rde_filter_community(new->aspath,
-		    COMMUNITY_WELLKNOWN, COMMUNITY_NO_ADVERTISE)) {
-			up_generate_updates(peer, NULL, old);
+		case 0:
+			up_generate_updates(rules, peer, NULL, old);
+			return;
+		case -1:
 			return;
 		}
-		if (peer->conf.ebgp && rde_filter_community(new->aspath,
-		    COMMUNITY_WELLKNOWN, COMMUNITY_NO_EXPORT)) {
-			up_generate_updates(peer, NULL, old);
-			return;
-		}
-		if (peer->conf.ebgp && rde_filter_community(new->aspath,
-		    COMMUNITY_WELLKNOWN, COMMUNITY_NO_EXPSUBCONFED)) {
-			up_generate_updates(peer, NULL, old);
-			return;
-		}
-
-		/* copy attributes for output filter */
-		fasp = path_copy(new->aspath);
-
-		/*
-		 * apply default outgoing overrides,
-		 * actually only prepend-self
-		 */
-		rde_apply_set(fasp, &peer->conf.attrset, new->prefix->af,
-		    fasp->peer, DIR_DEFAULT_OUT);
 
 		pt_getaddr(new->prefix, &addr);
-		if (rde_filter(peer, fasp, &addr,
-		    new->prefix->prefixlen, DIR_OUT) == ACTION_DENY) {
-			path_put(fasp);
-			up_generate_updates(peer, NULL, old);
+		if (rde_filter(&asp, rules, peer, new->aspath, &addr,
+		    new->prefix->prefixlen, new->aspath->peer, DIR_OUT) ==
+		    ACTION_DENY) {
+			path_put(asp);
+			up_generate_updates(rules, peer, NULL, old);
 			return;
 		}
 
-		/*
-		 * Don't send messages back to originator
-		 * this is not specified in the RFC but seems logical.
-		 */
-		if ((attr = attr_optget(new->aspath,
-		    ATTR_ORIGINATOR_ID)) != NULL) {
-			if (memcmp(attr->data, &peer->remote_bgpid,
-			    sizeof(peer->remote_bgpid)) == 0) {
-				/* would cause loop don't send */
-				path_put(fasp);
-				return;
-			}
-		}
-
 		/* generate update */
-		p = calloc(1, sizeof(struct update_prefix));
-		if (p == NULL)
-			fatal("up_generate_updates");
-
-		a = calloc(1, sizeof(struct update_attr));
-		if (a == NULL)
-			fatal("up_generate_updates");
-
-		if (up_generate_attr(peer, a, fasp) == -1) {
-			log_warnx("generation of bgp path attributes failed");
-			free(a);
-			free(p);
-			return;
-		}
-
-		/*
-		 * use aspath_hash as attr_hash, this may be unoptimal
-		 * but currently I don't care.
-		 */
-		a->attr_hash = aspath_hash(fasp->aspath->data,
-		    fasp->aspath->len);
-		p->prefix = addr;
-		p->prefixlen = new->prefix->prefixlen;
-
-		/* no longer needed */
-		path_put(fasp);
-
-		if (up_add(peer, p, a) == -1)
-			log_warnx("queuing update failed.");
+		if (asp != NULL) {
+			up_generate(peer, asp, &addr, new->prefix->prefixlen);
+			path_put(asp);
+		} else
+			up_generate(peer, new->aspath, &addr,
+			    new->prefix->prefixlen);
 	}
 }
 
+/* send a default route to the specified peer */
 void
-up_generate_default(struct rde_peer *peer, sa_family_t af)
+up_generate_default(struct filter_head *rules, struct rde_peer *peer,
+    sa_family_t af)
 {
-	struct update_attr	*a;
-	struct update_prefix	*p;
-	struct rde_aspath	*asp;
+	struct rde_aspath	*asp, *fasp;
 	struct bgpd_addr	 addr;
 
+	if (peer->capa_received.mp_v4 == SAFI_NONE &&
+	    peer->capa_received.mp_v6 != SAFI_NONE &&
+	    af == AF_INET)
+		return;
+
+	if (peer->capa_received.mp_v6 == SAFI_NONE &&
+	    af == AF_INET6)
+		return;
+
 	asp = path_get();
 	asp->aspath = aspath_get(NULL, 0);
 	asp->origin = ORIGIN_IGP;
 	/* the other default values are OK, nexthop is once again NULL */
 
-	/* XXX apply default overrides. Not yet possible */
+	/*
+	 * XXX apply default overrides. Not yet possible, mainly a parse.y
+	 * problem.
+	 */
+	/* rde_apply_set(asp, set, af, NULL ???, DIR_IN); */
 
 	/* filter as usual */
 	bzero(&addr, sizeof(addr));
 	addr.af = af;
-	if (rde_filter(peer, asp, &addr, 0, DIR_OUT) == ACTION_DENY) {
+
+	if (rde_filter(&fasp, rules, peer, asp, &addr, 0, NULL, DIR_OUT) ==
+	    ACTION_DENY) {
+		path_put(fasp);
 		path_put(asp);
 		return;
 	}
 
 	/* generate update */
-	p = calloc(1, sizeof(struct update_prefix));
-	if (p == NULL)
-		fatal("up_generate_default");
-
-	a = calloc(1, sizeof(struct update_attr));
-	if (a == NULL)
-		fatal("up_generate_default");
-
-	if (up_generate_attr(peer, a, asp) == -1) {
-		log_warnx("generation of bgp path attributes failed");
-		free(a);
-		free(p);
-		return;
-	}
-
-	/*
-	 * use aspath_hash as attr_hash, this may be unoptimal
-	 * but currently I don't care.
-	 */
-	a->attr_hash = aspath_hash(asp->aspath->data, asp->aspath->len);
-	p->prefix = addr;
-	p->prefixlen = 0; /* default route */
+	if (fasp != NULL)
+		up_generate(peer, fasp, &addr, 0);
+	else
+		up_generate(peer, asp, &addr, 0);
 
 	/* no longer needed */
+	path_put(fasp);
 	path_put(asp);
-
-	if (up_add(peer, p, a) == -1)
-		log_warnx("queuing update failed.");
-
-}
-
-void
-up_dump_upcall(struct pt_entry *pt, void *ptr)
-{
-	struct rde_peer	*peer = ptr;
-
-	if (pt->active == NULL)
-		return;
-	up_generate_updates(peer, pt->active, NULL);
 }
 
 u_char	up_attr_buf[4096];
 
+/* only for IPv4 */
+in_addr_t
+up_get_nexthop(struct rde_peer *peer, struct rde_aspath *a)
+{
+	in_addr_t	mask;
+
+	/* nexthop, already network byte order */
+	if (a->flags & F_NEXTHOP_NOMODIFY) {
+		/* no modify flag set */
+		if (a->nexthop == NULL)
+			return (peer->local_v4_addr.v4.s_addr);
+		else
+			return (a->nexthop->exit_nexthop.v4.s_addr);
+	} else if (!peer->conf.ebgp) {
+		/*
+		 * If directly connected use peer->local_v4_addr
+		 * this is only true for announced networks.
+		 */
+		if (a->nexthop == NULL)
+			return (peer->local_v4_addr.v4.s_addr);
+		else if (a->nexthop->exit_nexthop.v4.s_addr ==
+		    peer->remote_addr.v4.s_addr)
+			/*
+			 * per rfc: if remote peer address is equal to
+			 * the nexthop set the nexthop to our local address.
+			 * This reduces the risk of routing loops.
+			 */
+			return (peer->local_v4_addr.v4.s_addr);
+		else
+			return (a->nexthop->exit_nexthop.v4.s_addr);
+	} else if (peer->conf.distance == 1) {
+		/* ebgp directly connected */
+		if (a->nexthop != NULL &&
+		    a->nexthop->flags & NEXTHOP_CONNECTED) {
+			mask = htonl(
+			    prefixlen2mask(a->nexthop->nexthop_netlen));
+			if ((peer->remote_addr.v4.s_addr & mask) ==
+			    (a->nexthop->nexthop_net.v4.s_addr & mask))
+				/* nexthop and peer are in the same net */
+				return (a->nexthop->exit_nexthop.v4.s_addr);
+			else
+				return (peer->local_v4_addr.v4.s_addr);
+		} else
+			return (peer->local_v4_addr.v4.s_addr);
+	} else
+		/* ebgp multihop */
+		/*
+		 * For ebgp multihop nh->flags should never have
+		 * NEXTHOP_CONNECTED set so it should be possible to unify the
+		 * two ebgp cases. But this is save and RFC compliant.
+		 */
+		return (peer->local_v4_addr.v4.s_addr);
+}
+
+int
+up_generate_mp_reach(struct rde_peer *peer, struct update_attr *upa,
+    struct rde_aspath *a, sa_family_t af)
+{
+	u_int16_t	tmp;
+
+	switch (af) {
+	case AF_INET6:
+		upa->mpattr_len = 21; /* AFI + SAFI + NH LEN + NH + SNPA LEN */
+		upa->mpattr = malloc(upa->mpattr_len);
+		if (upa->mpattr == NULL)
+			fatal("up_generate_mp_reach");
+		tmp = htons(AFI_IPv6);
+		memcpy(upa->mpattr, &tmp, sizeof(tmp));
+		upa->mpattr[2] = SAFI_UNICAST;
+		upa->mpattr[3] = sizeof(struct in6_addr);
+		upa->mpattr[20] = 0; /* SNPA always 0 */
+
+		/* nexthop dance see also up_get_nexthop() */
+		if (peer->conf.ebgp == 0) {
+			/* ibgp */
+			if (a->nexthop == NULL ||
+			    (a->nexthop->exit_nexthop.af == AF_INET6 &&
+			    memcmp(&a->nexthop->exit_nexthop.v6,
+			    &peer->remote_addr.v6, sizeof(struct in6_addr))))
+				memcpy(&upa->mpattr[4], &peer->local_v6_addr.v6,
+				    sizeof(struct in6_addr));
+			else
+				memcpy(&upa->mpattr[4],
+				    &a->nexthop->exit_nexthop.v6,
+				    sizeof(struct in6_addr));
+		} else if (peer->conf.distance == 1) {
+			/* ebgp directly connected */
+			if (a->nexthop != NULL &&
+			    a->nexthop->flags & NEXTHOP_CONNECTED)
+				if (prefix_compare(&peer->remote_addr,
+				    &a->nexthop->nexthop_net,
+				    a->nexthop->nexthop_netlen) == 0) {
+					/*
+					 * nexthop and peer are in the same
+					 * subnet
+					 */
+					memcpy(&upa->mpattr[4],
+					    &a->nexthop->exit_nexthop.v6,
+					    sizeof(struct in6_addr));
+					return (0);
+				}
+			memcpy(&upa->mpattr[4], &peer->local_v6_addr.v6,
+			    sizeof(struct in6_addr));
+		} else
+			/* ebgp multihop */
+			memcpy(&upa->mpattr[4], &peer->local_v6_addr.v6,
+			    sizeof(struct in6_addr));
+		return (0);
+	default:
+		break;
+	}
+	return (-1);
+}
+
 int
 up_generate_attr(struct rde_peer *peer, struct update_attr *upa,
-    struct rde_aspath *a)
+    struct rde_aspath *a, sa_family_t af)
 {
 	struct aspath	*path;
 	struct attr	*oa;
 	u_int32_t	 tmp32;
-	in_addr_t	 nexthop, mask;
-	int		 r;
+	in_addr_t	 nexthop;
+	int		 r, ismp = 0;
 	u_int16_t	 len = sizeof(up_attr_buf), wlen = 0;
+	u_int8_t	 l;
 
 	/* origin */
 	if ((r = attr_write(up_attr_buf + wlen, len, ATTR_WELL_KNOWN,
@@ -592,57 +645,18 @@ up_generate_attr(struct rde_peer *peer, struct update_attr *upa,
 	aspath_put(path);
 	wlen += r; len -= r;
 
-	/* nexthop, already network byte order */
-	if (a->flags & F_NEXTHOP_NOMODIFY) {
-		/* no modify flag set */
-		if (a->nexthop == NULL)
-			nexthop = peer->local_v4_addr.v4.s_addr;
-		else
-			nexthop = a->nexthop->exit_nexthop.v4.s_addr;
-	} else if (!peer->conf.ebgp) {
-		/*
-		 * If directly connected use peer->local_v4_addr
-		 * this is only true for announced networks.
-		 */
-		if (a->nexthop == NULL)
-			nexthop = peer->local_v4_addr.v4.s_addr;
-		else if (a->nexthop->exit_nexthop.v4.s_addr ==
-		    peer->remote_addr.v4.s_addr)
-			/*
-			 * per rfc: if remote peer address is equal to
-			 * the nexthop set the nexthop to our local address.
-			 * This reduces the risk of routing loops.
-			 */
-			nexthop = peer->local_v4_addr.v4.s_addr;
-		else
-			nexthop = a->nexthop->exit_nexthop.v4.s_addr;
-	} else if (peer->conf.distance == 1) {
-		/* ebgp directly connected */
-		if (a->nexthop != NULL &&
-		    a->nexthop->flags & NEXTHOP_CONNECTED) {
-			mask = 0xffffffff << (32 - a->nexthop->nexthop_netlen);
-			mask = htonl(mask);
-			if ((peer->remote_addr.v4.s_addr & mask) ==
-			    (a->nexthop->nexthop_net.v4.s_addr & mask))
-				/* nexthop and peer are in the same net */
-				nexthop = a->nexthop->exit_nexthop.v4.s_addr;
-			else
-				nexthop = peer->local_v4_addr.v4.s_addr;
-		} else
-			nexthop = peer->local_v4_addr.v4.s_addr;
-	} else
-		/* ebgp multihop */
-		/*
-		 * For ebgp multihop nh->flags should never have
-		 * NEXTHOP_CONNECTED set so it should be possible to unify the
-		 * two ebgp cases. But this is save and RFC compliant.
-		 */
-		nexthop = peer->local_v4_addr.v4.s_addr;
-
-	if ((r = attr_write(up_attr_buf + wlen, len, ATTR_WELL_KNOWN,
-	    ATTR_NEXTHOP, &nexthop, 4)) == -1)
-		return (-1);
-	wlen += r; len -= r;
+	switch (af) {
+	case AF_INET:
+		nexthop = up_get_nexthop(peer, a);
+		if ((r = attr_write(up_attr_buf + wlen, len, ATTR_WELL_KNOWN,
+		    ATTR_NEXTHOP, &nexthop, 4)) == -1)
+			return (-1);
+		wlen += r; len -= r;
+		break;
+	default:
+		ismp = 1;
+		break;
+	}
 
 	/*
 	 * The old MED from other peers MUST not be announced to others
@@ -674,7 +688,9 @@ up_generate_attr(struct rde_peer *peer, struct update_attr *upa,
 	 *  3. transitive known attrs: announce unmodified
 	 *  4. transitive unknown attrs: set partial bit and re-announce
 	 */
-	TAILQ_FOREACH(oa, &a->others, entry) {
+	for (l = 0; l < a->others_len; l++) {
+		if ((oa = a->others[l]) == NULL)
+			break;
 		switch (oa->type) {
 		case ATTR_ATOMIC_AGGREGATE:
 			if ((r = attr_write(up_attr_buf + wlen, len,
@@ -716,6 +732,11 @@ up_generate_attr(struct rde_peer *peer, struct update_attr *upa,
 		wlen += r; len -= r;
 	}
 
+	/* write mp attribute to different buffer */
+	if (ismp)
+		if (up_generate_mp_reach(peer, upa, a, AF_INET6) == -1)
+			return (-1);
+
 	/* the bgp path attributes are now stored in the global buf */
 	upa->attr = malloc(wlen);
 	if (upa->attr == NULL)
@@ -742,7 +763,8 @@ up_dump_prefix(u_char *buf, int len, struct uplist_prefix *prefix_head,
 			log_warnx("dequeuing update failed.");
 		TAILQ_REMOVE(upp->prefix_h, upp, prefix_l);
 		peer->up_pcnt--;
-		if (upp->prefix_h == &peer->withdraws)
+		if (upp->prefix_h == &peer->withdraws ||
+		    upp->prefix_h == &peer->withdraws6)
 			peer->up_wcnt--;
 		else
 			peer->up_nlricnt--;
@@ -769,6 +791,7 @@ up_dump_attrnlri(u_char *buf, int len, struct rde_peer *peer)
 				log_warnx("dequeuing update failed.");
 			TAILQ_REMOVE(&peer->updates, upa, attr_l);
 			free(upa->attr);
+			free(upa->mpattr);
 			free(upa);
 			peer->up_acnt--;
 		} else
@@ -803,6 +826,7 @@ up_dump_attrnlri(u_char *buf, int len, struct rde_peer *peer)
 			log_warnx("dequeuing update failed.");
 		TAILQ_REMOVE(&peer->updates, upa, attr_l);
 		free(upa->attr);
+		free(upa->mpattr);
 		free(upa);
 		peer->up_acnt--;
 	}
@@ -810,3 +834,145 @@ up_dump_attrnlri(u_char *buf, int len, struct rde_peer *peer)
 	return (wpos);
 }
 
+char *
+up_dump_mp_unreach(u_char *buf, u_int16_t *len, struct rde_peer *peer)
+{
+	int		wpos = 8;	/* reserve some space for header */
+	u_int16_t	datalen, tmp;
+	u_int16_t	attrlen = 2;	/* attribute header (without len) */
+	u_int8_t	flags = ATTR_OPTIONAL;
+
+	if (*len < wpos)
+		return (NULL);
+
+	datalen = up_dump_prefix(buf + wpos, *len - wpos,
+	    &peer->withdraws6, peer);
+	if (datalen == 0)
+		return (NULL);
+
+	if (datalen > 255) {
+		attrlen += 2 + datalen;
+		flags |= ATTR_EXTLEN;
+	} else {
+		attrlen += 1 + datalen;
+		buf++;
+	}
+	/* prepend header */
+	/* no IPv4 withdraws */
+	wpos = 0;
+	bzero(buf, sizeof(u_int16_t));
+	wpos += sizeof(u_int16_t);
+
+	/* attribute length */
+	tmp = htons(attrlen);
+	memcpy(buf + wpos, &tmp, sizeof(u_int16_t));
+	wpos += sizeof(u_int16_t);
+
+	/* mp attribute */
+	buf[wpos++] = flags;
+	buf[wpos++] = (u_char)ATTR_MP_UNREACH_NLRI;
+
+	if (datalen > 255) {
+		tmp = htons(datalen);
+		memcpy(buf + wpos, &tmp, sizeof(u_int16_t));
+		wpos += sizeof(u_int16_t);
+	} else
+		buf[wpos++] = (u_char)datalen;
+
+	/* total length includes the two 2-bytes length fields. */
+	*len = attrlen + 2 * sizeof(u_int16_t);
+
+	return (buf);
+}
+
+char *
+up_dump_mp_reach(u_char *buf, u_int16_t *len, struct rde_peer *peer)
+{
+	struct update_attr	*upa;
+	int			wpos;
+	u_int16_t		datalen, tmp;
+	u_int8_t		flags = ATTR_OPTIONAL;
+
+	/*
+	 * It is possible that a queued path attribute has no nlri prefix.
+	 * Ignore and remove those path attributes.
+	 */
+	while ((upa = TAILQ_FIRST(&peer->updates6)) != NULL)
+		if (TAILQ_EMPTY(&upa->prefix_h)) {
+			if (RB_REMOVE(uptree_attr, &peer->up_attrs,
+			    upa) == NULL)
+				log_warnx("dequeuing update failed.");
+			TAILQ_REMOVE(&peer->updates6, upa, attr_l);
+			free(upa->attr);
+			free(upa->mpattr);
+			free(upa);
+			peer->up_acnt--;
+		} else
+			break;
+
+	if (upa == NULL)
+		return (NULL);
+
+	/*
+	 * reserve space for attr len, the attributes, the
+	 * mp attribute and the attribute header
+	 */
+	wpos = 2 + 2 + upa->attr_len + 4 + upa->mpattr_len;
+	if (*len < wpos)
+		return (NULL);
+
+	datalen = up_dump_prefix(buf + wpos, *len - wpos,
+	    &upa->prefix_h, peer);
+	if (datalen == 0)
+		return (NULL);
+
+	if (upa->mpattr_len == 0 || upa->mpattr == NULL)
+		fatalx("mulitprotocol update without MP attrs");
+
+	datalen += upa->mpattr_len;
+	wpos -= upa->mpattr_len;
+	memcpy(buf + wpos, upa->mpattr, upa->mpattr_len);
+
+	if (datalen > 255) {
+		wpos -= 2;
+		tmp = htons(datalen);
+		memcpy(buf + wpos, &tmp, sizeof(tmp));
+		datalen += 4;
+		flags |= ATTR_EXTLEN;
+	} else {
+		buf++; wpos--; /* skip one byte */
+		buf[--wpos] = (u_char)datalen;
+		datalen += 3;
+	}
+	buf[--wpos] = (u_char)ATTR_MP_REACH_NLRI;
+	buf[--wpos] = flags;
+
+	datalen += upa->attr_len;
+	wpos -= upa->attr_len;
+	memcpy(buf + wpos, upa->attr, upa->attr_len);
+
+	if (wpos != 4)
+		fatalx("Grrr, mp_reach buffer fucked up");
+	*len = datalen + 4;
+
+	wpos -= 2;
+	tmp = htons(datalen);
+	memcpy(buf + wpos, &tmp, sizeof(tmp));
+
+	wpos -= 2;
+	bzero(buf + wpos, 2);
+
+	/* now check if all prefixes where written */
+	if (TAILQ_EMPTY(&upa->prefix_h)) {
+		if (RB_REMOVE(uptree_attr, &peer->up_attrs, upa) == NULL)
+			log_warnx("dequeuing update failed.");
+		TAILQ_REMOVE(&peer->updates6, upa, attr_l);
+		free(upa->attr);
+		free(upa->mpattr);
+		free(upa);
+		peer->up_acnt--;
+	}
+
+	return (buf);
+}
+
