@@ -1,7 +1,7 @@
-/*	$OpenBSD: session.c,v 1.213 2005/03/11 17:46:11 henning Exp $ */
+/*	$OpenBSD: session.c,v 1.244 2006/02/08 12:36:29 henning Exp $ */
 
 /*
- * Copyright (c) 2003, 2004 Henning Brauer <henning@openbsd.org>
+ * Copyright (c) 2003, 2004, 2005 Henning Brauer <henning@openbsd.org>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -46,7 +46,8 @@
 #define PFD_PIPE_MAIN		0
 #define PFD_PIPE_ROUTE		1
 #define	PFD_SOCK_CTL		2
-#define PFD_LISTENERS_START	3
+#define	PFD_SOCK_RCTL		3
+#define PFD_LISTENERS_START	4
 
 void	session_sighdlr(int);
 int	setup_listeners(u_int *);
@@ -91,7 +92,7 @@ struct bgpd_sysdep	 sysdep;
 struct peer		*npeers;
 volatile sig_atomic_t	 session_quit = 0;
 int			 pending_reconf = 0;
-int			 csock = -1;
+int			 csock = -1, rcsock = -1;
 u_int			 peer_cnt;
 struct imsgbuf		*ibuf_rde;
 struct imsgbuf		*ibuf_main;
@@ -193,7 +194,10 @@ session_main(struct bgpd_config *config, struct peer *cpeers,
 	}
 
 	/* control socket is outside chroot */
-	if ((csock = control_init()) == -1)
+	if ((csock = control_init(0, SOCKET_NAME)) == -1)
+		fatalx("control socket setup failed");
+	if (conf->rcsock != NULL &&
+	    (rcsock = control_init(1, conf->rcsock)) == -1)
 		fatalx("control socket setup failed");
 
 	if ((pw = getpwnam(BGPD_USER)) == NULL)
@@ -211,12 +215,10 @@ session_main(struct bgpd_config *config, struct peer *cpeers,
 		fatalx("pfkey setup failed");
 
 	if (setgroups(1, &pw->pw_gid) ||
-	    setegid(pw->pw_gid) || setgid(pw->pw_gid) ||
-	    seteuid(pw->pw_uid) || setuid(pw->pw_uid))
+	    setresgid(pw->pw_gid, pw->pw_gid, pw->pw_gid) ||
+	    setresuid(pw->pw_uid, pw->pw_uid, pw->pw_uid))
 		fatal("can't drop privileges");
 
-	endpwent();
-
 	listener_cnt = 0;
 	setup_listeners(&listener_cnt);
 
@@ -235,7 +237,8 @@ session_main(struct bgpd_config *config, struct peer *cpeers,
 	imsg_init(ibuf_rde, pipe_s2r[0]);
 	imsg_init(ibuf_main, pipe_m2s[1]);
 	TAILQ_INIT(&ctl_conns);
-	csock = control_listen();
+	control_listen(csock);
+	control_listen(rcsock);
 	LIST_INIT(&mrthead);
 	peer_cnt = 0;
 	ctl_cnt = 0;
@@ -250,6 +253,7 @@ session_main(struct bgpd_config *config, struct peer *cpeers,
 	/* network list is not used in the SE */
 	while ((net = TAILQ_FIRST(net_l)) != NULL) {
 		TAILQ_REMOVE(net_l, net, entry);
+		filterset_free(&net->net.attrset);
 		free(net);
 	}
 
@@ -352,6 +356,8 @@ session_main(struct bgpd_config *config, struct peer *cpeers,
 			pfd[PFD_PIPE_ROUTE].events |= POLLOUT;
 		pfd[PFD_SOCK_CTL].fd = csock;
 		pfd[PFD_SOCK_CTL].events = POLLIN;
+		pfd[PFD_SOCK_RCTL].fd = rcsock;
+		pfd[PFD_SOCK_RCTL].events = POLLIN;
 
 		nextaction = time(NULL) + 240;	/* loop every 240s at least */
 		i = PFD_LISTENERS_START;
@@ -381,6 +387,7 @@ session_main(struct bgpd_config *config, struct peer *cpeers,
 					p->IdleHoldTime =
 					    INTERVAL_IDLE_HOLD_INITIAL;
 					p->IdleHoldResetTimer = 0;
+					p->errcnt = 0;
 				} else
 					p->IdleHoldResetTimer =
 					    time(NULL) + p->IdleHoldTime;
@@ -463,7 +470,12 @@ session_main(struct bgpd_config *config, struct peer *cpeers,
 
 		if (nfds > 0 && pfd[PFD_SOCK_CTL].revents & POLLIN) {
 			nfds--;
-			ctl_cnt += control_accept(csock);
+			ctl_cnt += control_accept(csock, 0);
+		}
+
+		if (nfds > 0 && pfd[PFD_SOCK_RCTL].revents & POLLIN) {
+			nfds--;
+			ctl_cnt += control_accept(rcsock, 1);
 		}
 
 		for (j = PFD_LISTENERS_START; nfds > 0 && j < idx_listeners;
@@ -516,7 +528,8 @@ session_main(struct bgpd_config *config, struct peer *cpeers,
 	msgbuf_clear(&ibuf_main->w);
 	free(ibuf_main);
 
-	control_shutdown();
+	control_shutdown(csock);
+	control_shutdown(rcsock);
 	log_info("session engine exiting");
 	_exit(0);
 }
@@ -533,10 +546,6 @@ init_peer(struct peer *p)
 {
 	p->fd = p->wbuf.fd = -1;
 
-	memcpy(&p->capa.ann, &p->conf.capabilities, sizeof(p->capa.ann));
-	if (!p->conf.announce_capa)
-		session_capa_ann_none(p);
-
 	if (p->conf.if_depend[0])
 		imsg_compose(ibuf_main, IMSG_IFINFO, 0, 0, -1,
 		    p->conf.if_depend, sizeof(p->conf.if_depend));
@@ -546,7 +555,10 @@ init_peer(struct peer *p)
 	peer_cnt++;
 
 	change_state(p, STATE_IDLE, EVNT_NONE);
-	p->IdleHoldTimer = time(NULL);	/* start ASAP */
+	if (p->conf.down)
+		p->IdleHoldTimer = 0;		/* no autostart */
+	else
+		p->IdleHoldTimer = time(NULL);	/* start ASAP */
 }
 
 void
@@ -580,17 +592,22 @@ bgp_fsm(struct peer *peer, enum session_events event)
 				return;
 			}
 
+			peer->stats.last_sent_errcode = 0;
+			peer->stats.last_sent_suberr = 0;
+
 			if (!peer->depend_ok)
 				peer->ConnectRetryTimer = 0;
-			else if (peer->conf.passive || peer->conf.template) {
+			else if (peer->passive || peer->conf.passive ||
+			    peer->conf.template) {
 				change_state(peer, STATE_ACTIVE, event);
 				peer->ConnectRetryTimer = 0;
 			} else {
 				change_state(peer, STATE_CONNECT, event);
-				session_connect(peer);
 				peer->ConnectRetryTimer =
 				    time(NULL) + INTERVAL_CONNECTRETRY;
+				session_connect(peer);
 			}
+			peer->passive = 0;
 			break;
 		default:
 			/* ignore */
@@ -620,7 +637,7 @@ bgp_fsm(struct peer *peer, enum session_events event)
 			peer->ConnectRetryTimer =
 			    time(NULL) + INTERVAL_CONNECTRETRY;
 			session_connect(peer);
-		break;
+			break;
 		default:
 			change_state(peer, STATE_IDLE, event);
 			break;
@@ -856,6 +873,13 @@ change_state(struct peer *peer, enum session_state state,
 			    peer->IdleHoldTime < MAX_IDLE_HOLD/2)
 				peer->IdleHoldTime *= 2;
 		}
+		if (peer->state == STATE_NONE || peer->state == STATE_ESTABLISHED) {
+			/* initialize capability negotiation structures */
+			memcpy(&peer->capa.ann, &peer->conf.capabilities,
+			    sizeof(peer->capa.ann));
+			if (!peer->conf.announce_capa)
+				session_capa_ann_none(peer);
+		}
 		break;
 	case STATE_CONNECT:
 		break;
@@ -866,9 +890,7 @@ change_state(struct peer *peer, enum session_state state,
 	case STATE_OPENCONFIRM:
 		break;
 	case STATE_ESTABLISHED:
-		if (peer->IdleHoldTime > INTERVAL_IDLE_HOLD_INITIAL)
-			peer->IdleHoldResetTimer =
-			    time(NULL) + peer->IdleHoldTime;
+		peer->IdleHoldResetTimer = time(NULL) + peer->IdleHoldTime;
 		session_up(peer);
 		break;
 	default:		/* something seriously fucked */
@@ -909,6 +931,13 @@ session_accept(int listenfd)
 
 	p = getpeerbyip((struct sockaddr *)&cliaddr);
 
+	if (p != NULL && p->state == STATE_IDLE && p->errcnt < 2 &&
+	    p->ConnectRetryTimer > 0) {
+		/* fast reconnect after clear */
+		p->passive = 1;
+		bgp_fsm(p, EVNT_START);
+	}
+
 	if (p != NULL &&
 	    (p->state == STATE_CONNECT || p->state == STATE_ACTIVE)) {
 		if (p->fd != -1) {
@@ -1365,6 +1394,10 @@ session_notification(struct peer *peer, u_int8_t errcode, u_int8_t subcode,
 	struct mrt		*mrt;
 	ssize_t			 len;
 	int			 errs = 0;
+	u_int8_t		 null8 = 0;
+
+	if (peer->stats.last_sent_errcode)	/* some notifctn already sent */
+		return;
 
 	len = MSGSIZE_NOTIFICATION_MIN + datalen;
 
@@ -1380,7 +1413,10 @@ session_notification(struct peer *peer, u_int8_t errcode, u_int8_t subcode,
 	errs += buf_add(buf, &msg.len, sizeof(msg.len));
 	errs += buf_add(buf, &msg.type, sizeof(msg.type));
 	errs += buf_add(buf, &errcode, sizeof(errcode));
-	errs += buf_add(buf, &subcode, sizeof(subcode));
+	if (errcode == ERR_CEASE)
+		errs += buf_add(buf, &null8, sizeof(null8));
+	else
+		errs += buf_add(buf, &subcode, sizeof(subcode));
 
 	if (datalen > 0)
 		errs += buf_add(buf, data, datalen);
@@ -1406,7 +1442,10 @@ session_notification(struct peer *peer, u_int8_t errcode, u_int8_t subcode,
 		bgp_fsm(peer, EVNT_CON_FATAL);
 		return;
 	}
+
 	peer->stats.msg_sent_notification++;
+	peer->stats.last_sent_errcode = errcode;
+	peer->stats.last_sent_suberr = subcode;
 }
 
 int
@@ -1427,7 +1466,11 @@ session_dispatch_msg(struct pollfd *pfd, struct peer *p)
 				    &error, &len) == -1 || error) {
 					if (error)
 						errno = error;
-					log_peer_warn(&p->conf, "socket error");
+					if (errno != p->lasterr) {
+						log_peer_warn(&p->conf,
+						    "socket error");
+						p->lasterr = errno;
+					}
 					bgp_fsm(p, EVNT_CON_OPENFAIL);
 					return (1);
 				}
@@ -1471,78 +1514,78 @@ session_dispatch_msg(struct pollfd *pfd, struct peer *p)
 	if (pfd->revents & POLLIN) {
 		if ((n = read(p->fd, p->rbuf->buf + p->rbuf->wpos,
 		    sizeof(p->rbuf->buf) - p->rbuf->wpos)) == -1) {
-				if (errno != EINTR && errno != EAGAIN) {
-					log_peer_warn(&p->conf, "read error");
-					bgp_fsm(p, EVNT_CON_FATAL);
-				}
-				return (1);
-			}
-			if (n == 0) {	/* connection closed */
-				bgp_fsm(p, EVNT_CON_CLOSED);
-				return (1);
+			if (errno != EINTR && errno != EAGAIN) {
+				log_peer_warn(&p->conf, "read error");
+				bgp_fsm(p, EVNT_CON_FATAL);
 			}
+			return (1);
+		}
+		if (n == 0) {	/* connection closed */
+			bgp_fsm(p, EVNT_CON_CLOSED);
+			return (1);
+		}
 
-			rpos = 0;
-			av = p->rbuf->wpos + n;
-			p->stats.last_read = time(NULL);
+		rpos = 0;
+		av = p->rbuf->wpos + n;
+		p->stats.last_read = time(NULL);
 
-			/*
-			 * session might drop to IDLE -> buffers deallocated
-			 * we MUST check rbuf != NULL before use
-			 */
-			for (;;) {
-				if (rpos + MSGSIZE_HEADER > av)
-					break;
-				if (p->rbuf == NULL)
-					break;
-				if (parse_header(p, p->rbuf->buf + rpos,
-				    &msglen, &msgtype) == -1)
-					return (0);
-				if (rpos + msglen > av)
-					break;
-				p->rbuf->rptr = p->rbuf->buf + rpos;
-
-				switch (msgtype) {
-				case OPEN:
-					bgp_fsm(p, EVNT_RCVD_OPEN);
-					p->stats.msg_rcvd_open++;
-					break;
-				case UPDATE:
-					bgp_fsm(p, EVNT_RCVD_UPDATE);
-					p->stats.msg_rcvd_update++;
-					break;
-				case NOTIFICATION:
-					bgp_fsm(p, EVNT_RCVD_NOTIFICATION);
-					p->stats.msg_rcvd_notification++;
-					break;
-				case KEEPALIVE:
-					bgp_fsm(p, EVNT_RCVD_KEEPALIVE);
-					p->stats.msg_rcvd_keepalive++;
-					break;
-				case RREFRESH:
-					parse_refresh(p);
-					p->stats.msg_rcvd_rrefresh++;
-					break;
-				default:	/* cannot happen */
-					session_notification(p, ERR_HEADER,
-					    ERR_HDR_TYPE, &msgtype, 1);
-					log_warnx("received message with "
-					    "unknown type %u", msgtype);
-				}
-				rpos += msglen;
-				if (++processed > MSG_PROCESS_LIMIT)
-					break;
-			}
+		/*
+		 * session might drop to IDLE -> buffers deallocated
+		 * we MUST check rbuf != NULL before use
+		 */
+		for (;;) {
+			if (rpos + MSGSIZE_HEADER > av)
+				break;
 			if (p->rbuf == NULL)
-				return (1);
+				break;
+			if (parse_header(p, p->rbuf->buf + rpos, &msglen,
+			    &msgtype) == -1)
+				return (0);
+			if (rpos + msglen > av)
+				break;
+			p->rbuf->rptr = p->rbuf->buf + rpos;
 
-			if (rpos < av) {
-				left = av - rpos;
-				memcpy(&p->rbuf->buf, p->rbuf->buf + rpos,
-				    left);
-				p->rbuf->wpos = left;
-			} else
-				p->rbuf->wpos = 0;
+			switch (msgtype) {
+			case OPEN:
+				bgp_fsm(p, EVNT_RCVD_OPEN);
+				p->stats.msg_rcvd_open++;
+				break;
+			case UPDATE:
+				bgp_fsm(p, EVNT_RCVD_UPDATE);
+				p->stats.msg_rcvd_update++;
+				break;
+			case NOTIFICATION:
+				bgp_fsm(p, EVNT_RCVD_NOTIFICATION);
+				p->stats.msg_rcvd_notification++;
+				break;
+			case KEEPALIVE:
+				bgp_fsm(p, EVNT_RCVD_KEEPALIVE);
+				p->stats.msg_rcvd_keepalive++;
+				break;
+			case RREFRESH:
+				parse_refresh(p);
+				p->stats.msg_rcvd_rrefresh++;
+				break;
+			default:	/* cannot happen */
+				session_notification(p, ERR_HEADER,
+				    ERR_HDR_TYPE, &msgtype, 1);
+				log_warnx("received message with "
+				    "unknown type %u", msgtype);
+				bgp_fsm(p, EVNT_CON_FATAL);
+			}
+			rpos += msglen;
+			if (++processed > MSG_PROCESS_LIMIT)
+				break;
+		}
+		if (p->rbuf == NULL)
+			return (1);
+
+		if (rpos < av) {
+			left = av - rpos;
+			memcpy(&p->rbuf->buf, p->rbuf->buf + rpos, left);
+			p->rbuf->wpos = left;
+		} else
+			p->rbuf->wpos = 0;
 
 		return (1);
 	}
@@ -1579,6 +1622,7 @@ parse_header(struct peer *peer, u_char *data, u_int16_t *len, u_int8_t *type)
 		    "received message: illegal length: %u byte", *len);
 		session_notification(peer, ERR_HEADER, ERR_HDR_LEN,
 		    &olen, sizeof(olen));
+		bgp_fsm(peer, EVNT_CON_FATAL);
 		return (-1);
 	}
 
@@ -1589,6 +1633,7 @@ parse_header(struct peer *peer, u_char *data, u_int16_t *len, u_int8_t *type)
 			    "received OPEN: illegal len: %u byte", *len);
 			session_notification(peer, ERR_HEADER, ERR_HDR_LEN,
 			    &olen, sizeof(olen));
+			bgp_fsm(peer, EVNT_CON_FATAL);
 			return (-1);
 		}
 		break;
@@ -1599,6 +1644,7 @@ parse_header(struct peer *peer, u_char *data, u_int16_t *len, u_int8_t *type)
 			    *len);
 			session_notification(peer, ERR_HEADER, ERR_HDR_LEN,
 			    &olen, sizeof(olen));
+			bgp_fsm(peer, EVNT_CON_FATAL);
 			return (-1);
 		}
 		break;
@@ -1608,6 +1654,7 @@ parse_header(struct peer *peer, u_char *data, u_int16_t *len, u_int8_t *type)
 			    "received UPDATE: illegal len: %u byte", *len);
 			session_notification(peer, ERR_HEADER, ERR_HDR_LEN,
 			    &olen, sizeof(olen));
+			bgp_fsm(peer, EVNT_CON_FATAL);
 			return (-1);
 		}
 		break;
@@ -1617,6 +1664,7 @@ parse_header(struct peer *peer, u_char *data, u_int16_t *len, u_int8_t *type)
 			    "received KEEPALIVE: illegal len: %u byte", *len);
 			session_notification(peer, ERR_HEADER, ERR_HDR_LEN,
 			    &olen, sizeof(olen));
+			bgp_fsm(peer, EVNT_CON_FATAL);
 			return (-1);
 		}
 		break;
@@ -1626,6 +1674,7 @@ parse_header(struct peer *peer, u_char *data, u_int16_t *len, u_int8_t *type)
 			    "received RREFRESH: illegal len: %u byte", *len);
 			session_notification(peer, ERR_HEADER, ERR_HDR_LEN,
 			    &olen, sizeof(olen));
+			bgp_fsm(peer, EVNT_CON_FATAL);
 			return (-1);
 		}
 		break;
@@ -1634,6 +1683,7 @@ parse_header(struct peer *peer, u_char *data, u_int16_t *len, u_int8_t *type)
 		    "received msg with unknown type %u", *type);
 		session_notification(peer, ERR_HEADER, ERR_HDR_TYPE,
 		    type, 1);
+		bgp_fsm(peer, EVNT_CON_FATAL);
 		return (-1);
 	}
 	LIST_FOREACH(mrt, &mrthead, entry) {
@@ -1737,7 +1787,7 @@ parse_open(struct peer *peer)
 	memcpy(&optparamlen, p, sizeof(optparamlen));
 	p += sizeof(optparamlen);
 
-	if (optparamlen > msglen - MSGSIZE_OPEN_MIN) {
+	if (optparamlen != msglen - MSGSIZE_OPEN_MIN) {
 			log_peer_warnx(&peer->conf,
 			    "corrupt OPEN message received: length mismatch");
 			session_notification(peer, ERR_OPEN, 0, NULL, 0);
@@ -1891,6 +1941,7 @@ parse_notification(struct peer *peer)
 	datalen -= sizeof(subcode);
 
 	log_notification(peer, errcode, subcode, p, datalen);
+	peer->errcnt++;
 
 	if (errcode == ERR_OPEN && subcode == ERR_OPEN_CAPA) {
 		if (datalen == 0) {	/* zebra likes to send those.. humbug */
@@ -1915,8 +1966,9 @@ parse_notification(struct peer *peer)
 			datalen -= sizeof(capa_len);
 			if (datalen < capa_len) {
 				log_peer_warnx(&peer->conf,
-				    "parse_notification: capa_len %u exceeds"
-				    "remaining msg length", capa_len);
+				    "parse_notification: capa_len %u exceeds "
+				    "remaining msg length %u", capa_len,
+				    datalen);
 				return (-1);
 			}
 			p += capa_len;
@@ -2037,8 +2089,6 @@ parse_capabilities(struct peer *peer, u_char *d, u_int16_t dlen)
 	return (0);
 }
 
-struct filter_set_head	*session_set;
-
 void
 session_dispatch_imsg(struct imsgbuf *ibuf, int idx, u_int *listener_cnt)
 {
@@ -2049,7 +2099,6 @@ session_dispatch_imsg(struct imsgbuf *ibuf, int idx, u_int *listener_cnt)
 	struct peer		*p, *next;
 	struct listen_addr	*la, *nla;
 	struct kif		*kif;
-	struct filter_set	*s;
 	u_char			*data;
 	enum reconf_action	 reconf;
 	int			 n, depend_ok;
@@ -2101,25 +2150,29 @@ session_dispatch_imsg(struct imsgbuf *ibuf, int idx, u_int *listener_cnt)
 				reconf = RECONF_KEEP;
 
 			memcpy(&p->conf, pconf, sizeof(struct peer_config));
-			SIMPLEQ_INIT(&p->conf.attrset);
-			session_set = &p->conf.attrset;
 			p->conf.reconf_action = reconf;
 			break;
 		case IMSG_RECONF_LISTENER:
 			if (idx != PFD_PIPE_MAIN)
 				fatalx("reconf request not from parent");
+			if (nconf == NULL)
+				fatalx("IMSG_RECONF_LISTENER but no config");
 			nla = imsg.data;
-			TAILQ_FOREACH(la, conf->listen_addrs, entry) {
+			TAILQ_FOREACH(la, conf->listen_addrs, entry)
 				if (!la_cmp(la, nla))
 					break;
-			}
-
-			if ((nla->fd = imsg_get_fd(ibuf)) == -1)
-				log_warnx("expected to receive fd for %s "
-				    "but didn't receive any",
-				    log_sockaddr((struct sockaddr *)&la->sa));
 
 			if (la == NULL) {
+				if (nla->reconf != RECONF_REINIT)
+					fatalx("king bula sez: "
+					    "expected REINIT");
+
+				if ((nla->fd = imsg_get_fd(ibuf)) == -1)
+					log_warnx("expected to receive fd for "
+					    "%s but didn't receive any",
+					    log_sockaddr((struct sockaddr *)
+					    &nla->sa));
+
 				la = calloc(1, sizeof(struct listen_addr));
 				if (la == NULL)
 					fatal(NULL);
@@ -2130,9 +2183,11 @@ session_dispatch_imsg(struct imsgbuf *ibuf, int idx, u_int *listener_cnt)
 				TAILQ_INSERT_TAIL(nconf->listen_addrs, la,
 				    entry);
 			} else {
+				if (nla->reconf != RECONF_KEEP)
+					fatalx("king bula sez: expected KEEP");
 				la->reconf = RECONF_KEEP;
-				close(nla->fd);
 			}
+
 			break;
 		case IMSG_RECONF_DONE:
 			if (idx != PFD_PIPE_MAIN)
@@ -2156,12 +2211,6 @@ session_dispatch_imsg(struct imsgbuf *ibuf, int idx, u_int *listener_cnt)
 				    !p->conf.cloned)
 					p->conf.reconf_action = RECONF_DELETE;
 
-			/* if there are no new listeners, keep default ones */
-			if (TAILQ_EMPTY(nconf->listen_addrs))
-				TAILQ_FOREACH(la, conf->listen_addrs, entry)
-					if (la->flags & DEFAULT_LISTENER)
-						la->reconf = RECONF_KEEP;
-
 			/* delete old listeners */
 			for (la = TAILQ_FIRST(conf->listen_addrs); la != NULL;
 			    la = nla) {
@@ -2189,22 +2238,9 @@ session_dispatch_imsg(struct imsgbuf *ibuf, int idx, u_int *listener_cnt)
 			free(nconf->listen_addrs);
 			free(nconf);
 			nconf = NULL;
-			session_set = NULL;
 			pending_reconf = 0;
 			log_info("SE reconfigured");
 			break;
-		case IMSG_FILTER_SET:
-			if (idx != PFD_PIPE_MAIN)
-				fatalx("reconf request not from parent");
-			if (session_set == NULL) {
-				log_warnx("IMSG_FILTER_SET unexpected");
-				break;
-			}
-			if ((s = malloc(sizeof(struct filter_set))) == NULL)
-				fatal(NULL);
-			memcpy(s, imsg.data, sizeof(struct filter_set));
-			SIMPLEQ_INSERT_TAIL(session_set, s, entry);
-			break;
 		case IMSG_IFINFO:
 			if (idx != PFD_PIPE_MAIN)
 				fatalx("IFINFO message not from parent");
@@ -2272,6 +2308,7 @@ session_dispatch_imsg(struct imsgbuf *ibuf, int idx, u_int *listener_cnt)
 			}
 			break;
 		case IMSG_CTL_KROUTE:
+		case IMSG_CTL_KROUTE6:
 		case IMSG_CTL_KROUTE_ADDR:
 		case IMSG_CTL_SHOW_NEXTHOP:
 		case IMSG_CTL_SHOW_INTERFACE:
@@ -2281,7 +2318,9 @@ session_dispatch_imsg(struct imsgbuf *ibuf, int idx, u_int *listener_cnt)
 			break;
 		case IMSG_CTL_SHOW_RIB:
 		case IMSG_CTL_SHOW_RIB_PREFIX:
+		case IMSG_CTL_SHOW_RIB_MEM:
 		case IMSG_CTL_SHOW_NETWORK:
+		case IMSG_CTL_SHOW_NETWORK6:
 		case IMSG_CTL_SHOW_NEIGHBOR:
 			if (idx != PFD_PIPE_ROUTE)
 				fatalx("ctl rib request not from RDE");
@@ -2323,6 +2362,21 @@ session_dispatch_imsg(struct imsgbuf *ibuf, int idx, u_int *listener_cnt)
 
 			session_notification(p, errcode, subcode,
 			    data, imsg.hdr.len - IMSG_HEADER_SIZE - 2);
+			switch (errcode) {
+			case ERR_CEASE:
+				switch (subcode) {
+				case ERR_CEASE_MAX_PREFIX:
+					bgp_fsm(p, EVNT_STOP);
+					break;
+				default:
+					bgp_fsm(p, EVNT_CON_FATAL);
+					break;
+				}
+				break;
+			default:
+				bgp_fsm(p, EVNT_CON_FATAL);
+				break;
+			}
 			break;
 		default:
 			break;
@@ -2480,7 +2534,7 @@ session_match_mask(struct peer *p, struct sockaddr *ip)
 	struct in6_addr	 mask;
 
 	if (p->conf.remote_addr.af == AF_INET) {
-		v4mask = htonl(0xffffffff << (32 - p->conf.remote_masklen));
+		v4mask = htonl(prefixlen2mask(p->conf.remote_masklen));
 		if (p->conf.remote_addr.v4.s_addr ==
 		    ((((struct sockaddr_in *)ip)->sin_addr.s_addr) & v4mask))
 			return (1);
@@ -2535,18 +2589,11 @@ void
 session_up(struct peer *p)
 {
 	struct session_up	 sup;
-	struct filter_set	*s;
 
 	if (imsg_compose(ibuf_rde, IMSG_SESSION_ADD, p->conf.id, 0, -1,
 	    &p->conf, sizeof(p->conf)) == -1)
 		fatalx("imsg_compose error");
 
-	SIMPLEQ_FOREACH(s, &p->conf.attrset, entry) {
-		if (imsg_compose(ibuf_rde, IMSG_FILTER_SET, p->conf.id, 0, -1,
-		    s, sizeof(struct filter_set)) == -1)
-			fatalx("imsg_compose error");
-	}
-
 	switch (p->sa_local.ss_family) {
 	case AF_INET:
 		sup.local_addr.af = AF_INET;
