@@ -1,4 +1,4 @@
-/*	$OpenBSD: session.h,v 1.75 2004/12/23 17:24:03 henning Exp $ */
+/*	$OpenBSD: session.h,v 1.81 2006/01/24 10:03:44 henning Exp $ */
 
 /*
  * Copyright (c) 2003, 2004 Henning Brauer <henning@openbsd.org>
@@ -35,6 +35,7 @@
 #define	MSGSIZE_KEEPALIVE		MSGSIZE_HEADER
 #define	MSGSIZE_RREFRESH		MSGSIZE_HEADER + 4
 #define	MSG_PROCESS_LIMIT		25
+#define	SESSION_CLEAR_DELAY		5
 
 enum session_state {
 	STATE_NONE,
@@ -105,7 +106,7 @@ enum capa_codes {
 };
 
 struct msg_header {
-	u_char			 marker[16];
+	u_char			 marker[MSGSIZE_HEADER_MARKER];
 	u_int16_t		 len;
 	u_int8_t		 type;
 };
@@ -127,6 +128,7 @@ struct bgpd_sysdep {
 struct ctl_conn {
 	TAILQ_ENTRY(ctl_conn)	entry;
 	struct imsgbuf		ibuf;
+	int			restricted;
 };
 
 TAILQ_HEAD(ctl_conns, ctl_conn)	ctl_conns;
@@ -145,6 +147,8 @@ struct peer_stats {
 	time_t			 last_updown;
 	time_t			 last_read;
 	u_int32_t		 prefix_cnt;
+	u_int8_t		 last_sent_errcode;
+	u_int8_t		 last_sent_suberr;
 };
 
 struct peer {
@@ -166,6 +170,7 @@ struct peer {
 	time_t			 IdleHoldResetTimer;
 	int			 fd;
 	int			 lasterr;
+	u_int			 errcnt;
 	u_int			 IdleHoldTime;
 	u_int32_t		 remote_bgpid;
 	enum session_state	 state;
@@ -173,6 +178,7 @@ struct peer {
 	u_int16_t		 holdtime;
 	u_int8_t		 auth_established;
 	u_int8_t		 depend_ok;
+	u_int8_t		 passive;
 };
 
 struct peer	*peers;
@@ -194,7 +200,6 @@ void		 log_statechange(struct peer *, enum session_state,
 void		 log_notification(const struct peer *, u_int8_t, u_int8_t,
 		    u_char *, u_int16_t);
 void		 log_conn_attempt(const struct peer *, struct sockaddr *);
-const char *	 log_sockaddr(struct sockaddr *);
 
 /* parse.y */
 int	 parse_config(char *, struct bgpd_config *, struct mrt_head *,
@@ -210,10 +215,11 @@ pid_t	 rde_main(struct bgpd_config *, struct peer *, struct network_head *,
 	    struct filter_head *, struct mrt_head *, int[2], int[2], int[2]);
 
 /* control.c */
-int	control_listen(void);
-void	control_shutdown(void);
+int	control_init(int, char *);
+int	control_listen(int);
+void	control_shutdown(int);
 int	control_dispatch_msg(struct pollfd *, u_int *);
-int	control_accept(int);
+int	control_accept(int, int);
 
 /* pfkey.c */
 int	pfkey_establish(struct peer *);
